# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `joint.py`."""

from absl.testing import absltest
from absl.testing import parameterized
import chex
from distrax._src.distributions.categorical import Categorical
from distrax._src.distributions.joint import _leaves_up_to_distribution
from distrax._src.distributions.joint import _map_up_to_distribution
from distrax._src.distributions.joint import Joint
from distrax._src.distributions.mvn_diag import MultivariateNormalDiag
from distrax._src.distributions.normal import Normal
import jax
import jax.numpy as jnp
import numpy as np


def _make_nested_distributions_and_inputs(batch_shape=None, shift=0.0):
  distributions = dict(
      categoricals=[
          Categorical(np.array([0 + shift, 1, 2])),
          Categorical(np.array([2 + shift, 0, 1])),
      ],
      normals=(
          Normal(np.array(1) + shift, np.array(2) + shift),
          Normal(np.array(2) + shift, np.array(1) + shift),
          Normal(np.array(5) + shift, np.array(4) + shift),
      ),
      multivariate=MultivariateNormalDiag(
          np.array([5, 6, 7]) + shift, np.array([3, 3, 3]) + shift
      ),
  )

  inputs = dict(
      categoricals=[
          np.array(0),
          np.array(1),
      ],
      normals=(
          np.array(2.5),
          np.array(3.1),
          np.array(4.2),
      ),
      multivariate=np.array([5.1, 5.2, 5.3]),
  )

  if isinstance(batch_shape, tuple):

    def _add_batch_shape_to_tensor(x):
      for dim in reversed(batch_shape):
        x = jnp.repeat(x[None, ...], dim, axis=0)
      return x

    def _add_batch_shape_to_distribution(d):
      if isinstance(d, Categorical):
        return Categorical(_add_batch_shape_to_tensor(d.logits_parameter()))
      elif isinstance(d, Normal):
        return Normal(
            _add_batch_shape_to_tensor(d.loc),
            _add_batch_shape_to_tensor(d.scale),
        )
      elif isinstance(d, MultivariateNormalDiag):
        return MultivariateNormalDiag(
            _add_batch_shape_to_tensor(d.loc),
            _add_batch_shape_to_tensor(d.scale_diag),
        )

    distributions = _map_up_to_distribution(
        _add_batch_shape_to_distribution, distributions
    )
    inputs = jax.tree.map(_add_batch_shape_to_tensor, inputs)

  return distributions, inputs


class JointTest(parameterized.TestCase):

  @chex.all_variants
  @parameterized.named_parameters(
      ('categorical', Categorical, (np.array([0, 1, 2]),), np.array(0)),
      (
          'normal',
          Normal,
          (np.array([0, 1, 2]), np.array([1, 2, 3])),
          np.array([0, 0, 0]),
      ),
      (
          'mvn',
          MultivariateNormalDiag,
          (np.array([[0, 1, 2], [3, 4, 5]]), np.array([[1, 2, 3], [4, 5, 6]])),
          np.array([[0, 0, 0], [0, 0, 0]]),
      ),
  )
  def test_single_distribution(self, fn, params, x):
    dist = fn(*params)
    joint = Joint(dist)

    key = jax.random.PRNGKey(0)
    (subkey,) = jax.random.split(key, 1)

    with self.subTest('sample'):
      actual = self.variant(joint.sample)(seed=key)
      expected = dist.sample(seed=subkey)
      np.testing.assert_allclose(actual, expected, rtol=1e-6)

    with self.subTest('log_prob'):
      actual = self.variant(joint.log_prob)(x)
      expected = dist.log_prob(x)
      np.testing.assert_allclose(actual, expected, rtol=3e-5)

    with self.subTest('sample_and_log_prob'):
      actual_sample, actual_log_prob = self.variant(joint.sample_and_log_prob)(
          seed=key
      )
      expected_sample, expected_log_prob = dist.sample_and_log_prob(seed=subkey)
      np.testing.assert_allclose(actual_sample, expected_sample, rtol=3e-5)
      np.testing.assert_allclose(actual_log_prob, expected_log_prob, rtol=3e-5)

  @chex.all_variants
  def test_distribution_tuple(self):
    distributions = (
        Categorical(np.array([0, 1, 2])),
        MultivariateNormalDiag(np.array([1, 2, 3]), np.array([2, 3, 4])),
    )
    inputs = (np.array(0), np.array([0.1, 0.2, 0.3]))
    joint = Joint(distributions)

    key = jax.random.PRNGKey(0)
    subkeys = jax.random.split(key, 2)

    with self.subTest('sample'):
      actuals = self.variant(joint.sample)(seed=key)
      assert isinstance(actuals, tuple)
      for actual, dist, subkey in zip(actuals, distributions, subkeys):
        expected = dist.sample(seed=subkey)
        np.testing.assert_allclose(actual, expected, rtol=1e-6)

    with self.subTest('log_prob'):
      actual = self.variant(joint.log_prob)(inputs)
      log_probs = [dist.log_prob(x) for dist, x in zip(distributions, inputs)]
      expected = sum(log_probs)
      np.testing.assert_array_equal(actual, expected)

    with self.subTest('sample_and_log_prob'):
      actual_sample, actual_log_prob = self.variant(joint.sample_and_log_prob)(
          seed=key
      )
      assert isinstance(actual_sample, tuple)
      samples = []
      log_probs = []
      for dist, subkey in zip(distributions, subkeys):
        sample, log_prob = dist.sample_and_log_prob(seed=subkey)
        samples.append(sample)
        log_probs.append(log_prob)
      expected_sample = tuple(samples)
      expected_log_prob = sum(log_probs)
      for actual, expected in zip(actual_sample, expected_sample):
        np.testing.assert_allclose(actual, expected, rtol=1e-6)
      np.testing.assert_array_equal(actual_log_prob, expected_log_prob)

  @chex.all_variants
  def test_distribution_list(self):
    distributions = [
        Categorical(np.array([0, 1, 2])),
        MultivariateNormalDiag(np.array([1, 2, 3]), np.array([2, 3, 4])),
    ]
    inputs = [np.array(0), np.array([0.1, 0.2, 0.3])]
    joint = Joint(distributions)

    key = jax.random.PRNGKey(0)
    subkeys = jax.random.split(key, 2)

    with self.subTest('sample'):
      actuals = self.variant(joint.sample)(seed=key)
      assert isinstance(actuals, list)
      for actual, dist, subkey in zip(actuals, distributions, subkeys):
        expected = dist.sample(seed=subkey)
        np.testing.assert_allclose(actual, expected, rtol=1e-6)

    with self.subTest('log_prob'):
      actual = self.variant(joint.log_prob)(inputs)
      log_probs = [dist.log_prob(x) for dist, x in zip(distributions, inputs)]
      expected = sum(log_probs)
      np.testing.assert_array_equal(actual, expected)

    with self.subTest('sample_and_log_prob'):
      actual_sample, actual_log_prob = self.variant(joint.sample_and_log_prob)(
          seed=key
      )
      assert isinstance(actual_sample, list)
      expected_sample = []
      log_probs = []
      for dist, subkey in zip(distributions, subkeys):
        sample, log_prob = dist.sample_and_log_prob(seed=subkey)
        expected_sample.append(sample)
        log_probs.append(log_prob)
      expected_log_prob = sum(log_probs)
      for actual, expected in zip(actual_sample, expected_sample):
        np.testing.assert_allclose(actual, expected, rtol=1e-6)
      np.testing.assert_array_equal(actual_log_prob, expected_log_prob)

  @chex.all_variants
  def test_distributions_with_batch_shape(self):
    distributions = [
        Categorical(np.array([[0, 1, 2], [3, 4, 5]])),
        MultivariateNormalDiag(
            np.array([[0, 1, 2, 3, 4], [2, 3, 4, 5, 6]]),
            np.array([[1, 2, 3, 5, 6], [2, 3, 4, 5, 6]]),
        ),
    ]
    inputs = [np.array([0, 1]), np.zeros((2, 5))]
    joint = Joint(distributions)
    assert joint.batch_shape == distributions[0].batch_shape
    assert joint.batch_shape == distributions[1].batch_shape

    key = jax.random.PRNGKey(0)
    subkeys = jax.random.split(key, 2)

    with self.subTest('sample'):
      actuals = self.variant(joint.sample)(seed=key)
      assert isinstance(actuals, list)
      assert actuals[0].shape == (2,)
      assert actuals[1].shape == (2, 5)
      for actual, dist, subkey in zip(actuals, distributions, subkeys):
        expected = dist.sample(seed=subkey)
        np.testing.assert_allclose(actual, expected, rtol=1e-6)

    with self.subTest('log_prob'):
      actual = self.variant(joint.log_prob)(inputs)
      assert actual.shape == (2,)
      log_probs = [dist.log_prob(x) for dist, x in zip(distributions, inputs)]
      expected = sum(log_probs)
      np.testing.assert_allclose(actual, expected, rtol=1e-6)

    with self.subTest('sample_and_log_prob'):
      actual_sample, actual_log_prob = self.variant(joint.sample_and_log_prob)(
          seed=key
      )
      assert isinstance(actual_sample, list)
      assert actual_sample[0].shape == (2,)
      assert actual_sample[1].shape == (2, 5)
      assert actual_log_prob.shape == (2,)
      expected_sample = []
      log_probs = []
      for dist, subkey in zip(distributions, subkeys):
        sample, log_prob = dist.sample_and_log_prob(seed=subkey)
        expected_sample.append(sample)
        log_probs.append(log_prob)
      expected_log_prob = sum(log_probs)
      for actual, expected in zip(actual_sample, expected_sample):
        np.testing.assert_allclose(actual, expected, rtol=1e-6)
      np.testing.assert_allclose(actual_log_prob, expected_log_prob, rtol=1e-6)

  @chex.all_variants
  def test_nested_distributions(self):
    distributions, inputs = _make_nested_distributions_and_inputs()
    joint = Joint(distributions)

    key = jax.random.PRNGKey(0)
    subkeys = jax.random.split(key, 6)

    with self.subTest('sample'):
      actuals = self.variant(joint.sample)(seed=key)
      assert isinstance(actuals, dict)
      assert isinstance(actuals['categoricals'], list)
      assert isinstance(actuals['normals'], tuple)
      assert isinstance(actuals['multivariate'], jnp.ndarray)

      flat_actuals = jax.tree.leaves(actuals)
      flat_dists = _leaves_up_to_distribution(distributions)
      for actual, dist, subkey in zip(flat_actuals, flat_dists, subkeys):
        expected = dist.sample(seed=subkey)
        np.testing.assert_allclose(actual, expected, rtol=1e-6)

    with self.subTest('log_prob'):
      actual = self.variant(joint.log_prob)(inputs)
      flat_dists = _leaves_up_to_distribution(distributions)
      flat_inputs = jax.tree.leaves(inputs)
      log_probs = [dist.log_prob(x) for dist, x in zip(flat_dists, flat_inputs)]
      expected = sum(log_probs)
      np.testing.assert_array_equal(actual, expected)

    with self.subTest('sample_and_log_prob'):
      actual_sample, actual_log_prob = self.variant(joint.sample_and_log_prob)(
          seed=key
      )
      assert isinstance(actual_sample, dict)
      assert isinstance(actual_sample['categoricals'], list)
      assert isinstance(actual_sample['normals'], tuple)
      assert isinstance(actual_sample['multivariate'], jnp.ndarray)

      expected_sample = []
      log_probs = []
      flat_dists = _leaves_up_to_distribution(distributions)
      for dist, subkey in zip(flat_dists, subkeys):
        sample, log_prob = dist.sample_and_log_prob(seed=subkey)
        expected_sample.append(sample)
        log_probs.append(log_prob)
      expected_log_prob = sum(log_probs)
      flat_actuals = jax.tree.leaves(actual_sample)
      for actual, expected in zip(flat_actuals, expected_sample):
        np.testing.assert_allclose(actual, expected, rtol=1e-6)
      np.testing.assert_allclose(actual_log_prob, expected_log_prob, rtol=1e-6)

  @chex.all_variants(with_pmap=False)
  def test_entropy(self):
    distributions, _ = _make_nested_distributions_and_inputs()
    joint = Joint(distributions)
    actual = self.variant(joint.entropy)()
    flat_dists = _leaves_up_to_distribution(distributions)
    expected = sum(dist.entropy() for dist in flat_dists)
    np.testing.assert_allclose(actual, expected, rtol=1e-6)

  @chex.all_variants(with_pmap=False)
  def test_mode(self):
    distributions, _ = _make_nested_distributions_and_inputs()
    joint = Joint(distributions)
    actual = self.variant(joint.mode)()
    expected = _map_up_to_distribution(lambda d: d.mode(), distributions)
    chex.assert_trees_all_equal(actual, expected)

  @chex.all_variants(with_pmap=False)
  def test_mean(self):
    distributions, _ = _make_nested_distributions_and_inputs()
    del distributions['categoricals']  # Mean is not defined for these.
    joint = Joint(distributions)
    actual = self.variant(joint.mean)()
    expected = _map_up_to_distribution(lambda d: d.mean(), distributions)
    chex.assert_trees_all_equal(actual, expected)

  @chex.all_variants(with_pmap=False)
  def test_median(self):
    distributions, _ = _make_nested_distributions_and_inputs()
    del distributions['categoricals']  # Median is not defined for these.
    joint = Joint(distributions)
    actual = self.variant(joint.median)()
    expected = _map_up_to_distribution(lambda d: d.median(), distributions)
    chex.assert_trees_all_equal(actual, expected)

  @chex.all_variants
  def test_kl_divergence(self):
    dists_a, _ = _make_nested_distributions_and_inputs(shift=0.0)
    dists_b, _ = _make_nested_distributions_and_inputs(shift=1.0)

    joint_a = Joint(dists_a)
    joint_b = Joint(dists_b)
    actual = self.variant(joint_a.kl_divergence)(joint_b)

    kls = []
    for dist_a, dist_b in zip(
        _leaves_up_to_distribution(dists_a), _leaves_up_to_distribution(dists_b)
    ):
      kls.append(dist_a.kl_divergence(dist_b))
    expected = sum(kls)

    np.testing.assert_allclose(actual, expected, rtol=1e-4)

  @chex.all_variants
  def test_log_cdf(self):
    distributions, inputs = _make_nested_distributions_and_inputs()
    joint = Joint(distributions)
    actual = self.variant(joint.log_cdf)(inputs)
    flat_dists = _leaves_up_to_distribution(distributions)
    flat_inputs = jax.tree.leaves(inputs)
    expected = sum(dist.log_cdf(x) for dist, x in zip(flat_dists, flat_inputs))
    np.testing.assert_allclose(actual, expected, rtol=1e-6)

  def test_distributions_property(self):
    distributions, _ = _make_nested_distributions_and_inputs()
    joint = Joint(distributions)
    chex.assert_trees_all_equal_structs(joint.distributions, distributions)

  def test_event_shape_property(self):
    distributions, _ = _make_nested_distributions_and_inputs()
    joint = Joint(distributions)
    all_event_shapes = joint.event_shape

    def _test_leaf(dist, event_shape):
      np.testing.assert_equal(dist.event_shape, event_shape)

    _map_up_to_distribution(_test_leaf, distributions, all_event_shapes)

  def test_dtype_property(self):
    distributions, _ = _make_nested_distributions_and_inputs()
    joint = Joint(distributions)
    all_dtypes = joint.dtype
    for dist, dtype in zip(
        _leaves_up_to_distribution(distributions), jax.tree.leaves(all_dtypes)
    ):
      np.testing.assert_equal(dist.dtype, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d batch first element', (10,), 0),
      ('1d batch last element', (10,), -1),
      ('1d batch first two elements', (10,), slice(0, 2)),
      ('1d batch first and third elements', (10,), slice(0, 4, 2)),
      ('2d batch first element', (10, 7), 0),
      ('2d batch last element', (10, 7), -1),
      ('2d batch first two elements', (10, 7), slice(0, 2)),
      ('2d batch first and third elements', (10, 7), slice(0, 4, 2)),
  )
  def test_indexing(self, batch_shape, index):
    distributions, inputs = _make_nested_distributions_and_inputs(
        batch_shape=batch_shape
    )
    inputs = jax.tree.map(lambda x: x[index], inputs)

    joint = Joint(distributions)
    joint_indexed = joint[index]

    key = jax.random.PRNGKey(0)
    subkeys = jax.random.split(key, 6)

    with self.subTest('batch shape'):
      for dist, indexed in zip(
          _leaves_up_to_distribution(distributions),
          _leaves_up_to_distribution(joint_indexed.distributions),
      ):
        assert dist[index].batch_shape == indexed.batch_shape

    with self.subTest('event shape'):
      for dist, indexed in zip(
          _leaves_up_to_distribution(distributions),
          _leaves_up_to_distribution(joint_indexed.distributions),
      ):
        assert dist[index].event_shape == indexed.event_shape

    with self.subTest('sample'):
      all_samples = self.variant(joint_indexed.sample)(seed=key)
      for dist, subkey, actual in zip(
          _leaves_up_to_distribution(distributions),
          subkeys,
          jax.tree.leaves(all_samples),
      ):
        expected = dist[index].sample(seed=subkey)
        np.testing.assert_allclose(actual, expected, rtol=1e-4)

    with self.subTest('sample_and_log_prob'):
      actual_samples, actual_log_probs = self.variant(
          joint_indexed.sample_and_log_prob
      )(seed=key)
      expected_outputs = [
          dist[index].sample_and_log_prob(seed=subkey)
          for dist, subkey in zip(
              _leaves_up_to_distribution(distributions), subkeys
          )
      ]
      expected_samples = [sample for sample, _ in expected_outputs]
      expected_log_probs = sum(lp for _, lp in expected_outputs)
      for actual, expected in zip(
          jax.tree.leaves(actual_samples), expected_samples
      ):
        np.testing.assert_allclose(actual, expected, rtol=1e-4)
      np.testing.assert_allclose(
          actual_log_probs, expected_log_probs, rtol=1e-6
      )

    with self.subTest('log_prob'):
      actual = self.variant(joint_indexed.log_prob)(inputs)
      expected = sum(
          dist[index].log_prob(x)
          for dist, x in zip(
              _leaves_up_to_distribution(distributions), jax.tree.leaves(inputs)
          )
      )
      np.testing.assert_allclose(actual, expected, rtol=1e-6)

  def test_raise_on_mismatched_batch_shape(self):
    distributions = dict(
        unbatched=Categorical(np.zeros((3,))),
        batched=Normal(np.zeros((3, 4, 5)), np.ones((3, 4, 5))),
    )

    with self.assertRaises(ValueError):
      Joint(distributions)

  @chex.all_variants
  def test_raise_on_incompatible_distributions_kl(self):
    distributions, _ = _make_nested_distributions_and_inputs()
    incompatible = dict(
        categoricals=distributions['normals'],
        normals=distributions['categoricals'],
        multivariate=distributions['multivariate'],
    )

    joint_a = Joint(distributions)
    joint_b = Joint(incompatible)

    with self.assertRaises(ValueError):
      self.variant(joint_a.kl_divergence)(joint_b)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper to adapt a Distrax distribution for use in TFP."""

from typing import Dict, Optional, Sequence, Tuple, Union

import chex
from distrax._src.distributions import distribution
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
ArrayNumpy = chex.ArrayNumpy
Distribution = distribution.Distribution
IntLike = distribution.IntLike
PRNGKey = chex.PRNGKey
tangent_spaces = tfp.experimental.tangent_spaces
TangentSpace = tangent_spaces.TangentSpace
EventT = distribution.EventT


def tfp_compatible_distribution(
    base_distribution: Distribution,
    name: Optional[str] = None) -> distribution.DistributionT:
  """Create a TFP-compatible distribution from a Distrax distribution.

  Given a Distrax distribution, return a wrapped distribution that behaves as a
  TFP distribution, to be used in TFP meta-distributions. In particular, the
  wrapped distribution implements the methods `allow_nan_stats`, `parameters`,
  `name`, `batch_shape_tensor`, `reparameterization_type` and
  `event_shape_tensor`; and the `batch_shape` and `event_shape` properties
  return a TFP `TensorShape`.

  The methods of the resulting distribution do not take a `name` argument,
  unlike their TFP equivalents.

  Args:
    base_distribution: A Distrax distribution.
    name: The distribution name.

  Returns:
    The wrapped distribution.
  """

  name_ = name

  class TFPCompatibleDistribution(base_distribution.__class__):
    """Class to wrap a Distrax distribution.

    The wrapped class dynamically inherits from `base_distribution`, so that
    computations involving the KL remain valid.
    """

    def __init__(self):
      pass

    def __getattr__(self, name: str):
      return getattr(base_distribution, name)

    def __getitem__(self, index):
      return tfp_compatible_distribution(base_distribution[index], name=name_)

    @property
    def allow_nan_stats(self) -> bool:
      """Proxy for the TFP property `allow_nan_stats`.

      It always returns True.
      """
      return True

    @property
    def batch_shape(self) -> tfp.tf2jax.TensorShape:
      """Returns a `TensorShape` with the `batch_shape` of the distribution."""
      return tfp.tf2jax.TensorShape(base_distribution.batch_shape)

    def batch_shape_tensor(self) -> Array:
      """See `Distribution.batch_shape`."""
      return jnp.array(base_distribution.batch_shape, dtype=jnp.int32)

    @property
    def event_shape(self) -> tfp.tf2jax.TensorShape:
      """Returns a `TensorShape` with the `event_shape` of the distribution."""
      return tfp.tf2jax.TensorShape(base_distribution.event_shape)

    def event_shape_tensor(self) -> ArrayNumpy:
      """See `Distribution.event_shape`."""
      return np.array(base_distribution.event_shape, dtype=jnp.int32)

    @property
    def experimental_shard_axis_names(self):
      return []

    @property
    def name(self) -> str:
      """See `Distribution.name`."""
      return name_ or f'TFPCompatible{base_distribution.name}'

    @property
    def reparameterization_type(self) -> tfd.ReparameterizationType:
      """Proxy for the TFP property `reparameterization_type`.

      It always returns `tfd.NOT_REPARAMETERIZED`.
      """
      return tfd.NOT_REPARAMETERIZED

    def _sample_n(self, key: PRNGKey, n: int) -> EventT:
      return base_distribution.sample(seed=key, sample_shape=(n,))

    def log_prob(self, value: EventT) -> Array:
      """See `Distribution.log_prob`."""
      return base_distribution.log_prob(value)

    @property
    def parameters(self) -> Dict[str, str]:
      """Returns a dictionary whose key 'name' maps to the distribution name."""
      return {'name': self.name}

    def sample(self,
               sample_shape: Union[IntLike, Sequence[IntLike]] = (),
               seed: Optional[Union[int, tfp.util.SeedStream]] = None,
               **unused_kwargs) -> EventT:
      """See `Distribution.sample`."""
      if not np.isscalar(sample_shape):
        sample_shape = tuple(sample_shape)
      return base_distribution.sample(sample_shape=sample_shape, seed=seed)

    def experimental_local_measure(
        self,
        value: Array,
        backward_compat: bool = True,
        **unused_kwargs) -> Tuple[Array, TangentSpace]:
      """Returns a log probability density together with a `TangentSpace`.

      See `tfd.distribution.Distribution.experimental_local_measure`, and
      Radul and Alexeev, AISTATS 2021, “The Base Measure Problem and its
      Solution”, https://arxiv.org/abs/2010.09647.

      Args:
        value: `float` or `double` `Array`.
        backward_compat: unused
        **unused_kwargs: unused

      Returns:
        log_prob: see `log_prob`.
        tangent_space: `tangent_spaces.FullSpace()`, representing R^n with the
          standard basis.
      """
      del backward_compat
      # We ignore the `backward_compat` flag and always act as though it's
      # true because Distrax bijectors and distributions need not follow the
      # base measure protocol from TFP.
      del unused_kwargs
      return self.log_prob(value), tangent_spaces.FullSpace()

  return TFPCompatibleDistribution()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Uniform distribution."""

import math
from typing import Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Uniform(distribution.Distribution):
  """Uniform distribution with `low` and `high` parameters."""

  equiv_tfp_cls = tfd.Uniform

  def __init__(self, low: Numeric = 0., high: Numeric = 1.):
    """Initializes a Uniform distribution.

    Args:
      low: Lower bound.
      high: Upper bound.
    """
    super().__init__()
    self._low = conversion.as_float_array(low)
    self._high = conversion.as_float_array(high)
    self._batch_shape = jax.lax.broadcast_shapes(
        self._low.shape, self._high.shape)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of the events."""
    return ()

  @property
  def low(self) -> Array:
    """Lower bound."""
    return jnp.broadcast_to(self._low, self.batch_shape)

  @property
  def high(self) -> Array:
    """Upper bound."""
    return jnp.broadcast_to(self._high, self.batch_shape)

  @property
  def range(self) -> Array:
    return self.high - self.low

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    return self._batch_shape

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    new_shape = (n,) + self.batch_shape
    uniform = jax.random.uniform(
        key=key, shape=new_shape, dtype=self.range.dtype, minval=0., maxval=1.)
    low = jnp.expand_dims(self._low, range(uniform.ndim - self._low.ndim))
    range_ = jnp.expand_dims(self.range, range(uniform.ndim - self.range.ndim))
    return low + range_ * uniform

  def _sample_n_and_log_prob(self, key: PRNGKey, n: int) -> Tuple[Array, Array]:
    """See `Distribution._sample_n_and_log_prob`."""
    samples = self._sample_n(key, n)
    log_prob = -jnp.log(self.range)
    log_prob = jnp.repeat(log_prob[None], n, axis=0)
    return samples, log_prob

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    return jnp.log(self.prob(value))

  def prob(self, value: EventT) -> Array:
    """See `Distribution.prob`."""
    return jnp.where(
        jnp.logical_or(value < self.low, value > self.high),
        jnp.zeros_like(value),
        jnp.ones_like(value) / self.range)

  def entropy(self) -> Array:
    """Calculates the entropy."""
    return jnp.log(self.range)

  def mean(self) -> Array:
    """Calculates the mean."""
    return (self.low + self.high) / 2.

  def variance(self) -> Array:
    """Calculates the variance."""
    return jnp.square(self.range) / 12.

  def stddev(self) -> Array:
    """Calculates the standard deviation."""
    return self.range / math.sqrt(12.)

  def median(self) -> Array:
    """Calculates the median."""
    return self.mean()

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    ones = jnp.ones_like(self.range)
    zeros = jnp.zeros_like(ones)
    result_if_not_big = jnp.where(
        value < self.low, zeros, (value - self.low) / self.range)
    return jnp.where(value > self.high, ones, result_if_not_big)

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return jnp.log(self.cdf(value))

  def __getitem__(self, index) -> 'Uniform':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Uniform(low=self.low[index], high=self.high[index])


def _kl_divergence_uniform_uniform(
    dist1: Union[Uniform, tfd.Uniform],
    dist2: Union[Uniform, tfd.Uniform],
    *unused_args, **unused_kwargs,
    ) -> Array:
  """Obtain the KL divergence `KL(dist1 || dist2)` between two Uniforms.

  Note that the KL divergence is infinite if the support of `dist1` is not a
  subset of the support of `dist2`.

  Args:
    dist1: A Uniform distribution.
    dist2: A Uniform distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  return jnp.where(
      jnp.logical_and(dist2.low <= dist1.low, dist1.high <= dist2.high),
      jnp.log(dist2.high - dist2.low) - jnp.log(dist1.high - dist1.low),
      jnp.inf)


# Register the KL functions with TFP.
tfd.RegisterKL(Uniform, Uniform)(_kl_divergence_uniform_uniform)
tfd.RegisterKL(Uniform, Uniform.equiv_tfp_cls)(_kl_divergence_uniform_uniform)
tfd.RegisterKL(Uniform.equiv_tfp_cls, Uniform)(_kl_divergence_uniform_uniform)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `lower_upper_triangular_affine.py`."""

import copy

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors.lower_upper_triangular_affine import LowerUpperTriangularAffine
from distrax._src.bijectors.tanh import Tanh
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


class LowerUpperTriangularAffineTest(parameterized.TestCase):

  def test_jacobian_is_constant_property(self):
    bijector = LowerUpperTriangularAffine(
        matrix=jnp.eye(4), bias=jnp.zeros((4,)))
    self.assertTrue(bijector.is_constant_jacobian)
    self.assertTrue(bijector.is_constant_log_det)

  def test_properties(self):
    bijector = LowerUpperTriangularAffine(
        matrix=jnp.array([[2., 3.], [4., 5.]]),
        bias=jnp.ones((2,)))
    lower = np.array([[1., 0.], [4., 1.]])
    upper = np.array([[2., 3.], [0., 5.]])
    np.testing.assert_allclose(bijector.lower, lower, atol=1e-6)
    np.testing.assert_allclose(bijector.upper, upper, atol=1e-6)
    np.testing.assert_allclose(bijector.matrix, lower @ upper, atol=1e-6)
    np.testing.assert_allclose(bijector.bias, np.ones((2,)), atol=1e-6)

  @parameterized.named_parameters(
      ('matrix is 0d', {'matrix': np.zeros(()), 'bias': np.zeros((4,))}),
      ('matrix is 1d', {'matrix': np.zeros((4,)), 'bias': np.zeros((4,))}),
      ('bias is 0d', {'matrix': np.zeros((4, 4)), 'bias': np.zeros(())}),
      ('matrix is not square',
       {'matrix': np.zeros((3, 4)), 'bias': np.zeros((4,))}),
      ('matrix and bias shapes do not agree',
       {'matrix': np.zeros((4, 4)), 'bias': np.zeros((3,))}),
  )
  def test_raises_with_invalid_parameters(self, bij_params):
    with self.assertRaises(ValueError):
      LowerUpperTriangularAffine(**bij_params)

  @chex.all_variants
  @parameterized.parameters(
      ((5,), (5,), (5,)),
      ((5,), (5,), ()),
      ((5,), (), (5,)),
      ((), (5,), (5,)),
      ((), (), (5,)),
      ((), (5,), ()),
      ((5,), (), ()),
  )
  def test_batched_parameters(self, matrix_batch_shape, bias_batch_shape,
                              input_batch_shape):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    matrix = jax.random.uniform(
        next(prng), matrix_batch_shape + (4, 4)) + jnp.eye(4)
    bias = jax.random.normal(next(prng), bias_batch_shape + (4,))
    bijector = LowerUpperTriangularAffine(matrix, bias)

    x = jax.random.normal(next(prng), input_batch_shape + (4,))
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(x)
    z, logdet_inv = self.variant(bijector.inverse_and_log_det)(x)

    output_batch_shape = jnp.broadcast_arrays(
        matrix[..., 0, 0], bias[..., 0], x[..., 0])[0].shape

    self.assertEqual(y.shape, output_batch_shape + (4,))
    self.assertEqual(z.shape, output_batch_shape + (4,))
    self.assertEqual(logdet_fwd.shape, output_batch_shape)
    self.assertEqual(logdet_inv.shape, output_batch_shape)

    matrix = jnp.broadcast_to(
        matrix, output_batch_shape + (4, 4)).reshape((-1, 4, 4))
    bias = jnp.broadcast_to(bias, output_batch_shape + (4,)).reshape((-1, 4))
    x = jnp.broadcast_to(x, output_batch_shape + (4,)).reshape((-1, 4))
    y = y.reshape((-1, 4))
    z = z.reshape((-1, 4))
    logdet_fwd = logdet_fwd.flatten()
    logdet_inv = logdet_inv.flatten()

    for i in range(np.prod(output_batch_shape)):
      bijector = LowerUpperTriangularAffine(matrix[i], bias[i])
      this_y, this_logdet_fwd = self.variant(bijector.forward_and_log_det)(x[i])
      this_z, this_logdet_inv = self.variant(bijector.inverse_and_log_det)(x[i])
      np.testing.assert_allclose(this_y, y[i], atol=9e-3)
      np.testing.assert_allclose(this_z, z[i], atol=7e-6)
      np.testing.assert_allclose(this_logdet_fwd, logdet_fwd[i], atol=1e-7)
      np.testing.assert_allclose(this_logdet_inv, logdet_inv[i], atol=7e-6)

  @chex.all_variants
  @parameterized.parameters(
      {'batch_shape': (), 'param_shape': ()},
      {'batch_shape': (3,), 'param_shape': (3,)},
      {'batch_shape': (2, 3), 'param_shape': (3,)},
  )
  def test_identity_initialization(self, batch_shape, param_shape):
    bijector = LowerUpperTriangularAffine(
        matrix=jnp.eye(4),
        bias=jnp.zeros(param_shape + (4,)))
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    x = jax.random.normal(next(prng), batch_shape + (4,))

    # Forward methods.
    y, logdet = self.variant(bijector.forward_and_log_det)(x)
    np.testing.assert_allclose(y, x, 8e-3)
    np.testing.assert_array_equal(logdet, jnp.zeros(batch_shape))

    # Inverse methods.
    x_rec, logdet = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_array_equal(x_rec, y)
    np.testing.assert_array_equal(logdet, jnp.zeros(batch_shape))

  @chex.all_variants
  @parameterized.parameters(
      {'batch_shape': (), 'param_shape': ()},
      {'batch_shape': (3,), 'param_shape': (3,)},
      {'batch_shape': (2, 3), 'param_shape': (3,)}
  )
  def test_inverse_methods(self, batch_shape, param_shape):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    matrix = jax.random.uniform(next(prng), param_shape + (4, 4)) + jnp.eye(4)
    bias = jax.random.normal(next(prng), param_shape + (4,))
    bijector = LowerUpperTriangularAffine(matrix, bias)
    x = jax.random.normal(next(prng), batch_shape + (4,))
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(x)
    x_rec, logdet_inv = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(x_rec, x, atol=9e-3)
    np.testing.assert_array_equal(logdet_fwd, -logdet_inv)

  @chex.all_variants
  def test_forward_jacobian_det(self):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    matrix = jax.random.uniform(next(prng), (4, 4)) + jnp.eye(4)
    bias = jax.random.normal(next(prng), (4,))
    bijector = LowerUpperTriangularAffine(matrix, bias)

    batched_x = jax.random.normal(next(prng), (10, 4))
    single_x = jax.random.normal(next(prng), (4,))
    batched_logdet = self.variant(bijector.forward_log_det_jacobian)(batched_x)

    jacobian_fn = jax.jacfwd(bijector.forward)
    logdet_numerical = jnp.linalg.slogdet(jacobian_fn(single_x))[1]
    for logdet in batched_logdet:
      np.testing.assert_allclose(logdet, logdet_numerical, atol=5e-3)

  @chex.all_variants
  def test_inverse_jacobian_det(self):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    matrix = jax.random.uniform(next(prng), (4, 4)) + jnp.eye(4)
    bias = jax.random.normal(next(prng), (4,))
    bijector = LowerUpperTriangularAffine(matrix, bias)

    batched_y = jax.random.normal(next(prng), (10, 4))
    single_y = jax.random.normal(next(prng), (4,))
    batched_logdet = self.variant(bijector.inverse_log_det_jacobian)(batched_y)

    jacobian_fn = jax.jacfwd(bijector.inverse)
    logdet_numerical = jnp.linalg.slogdet(jacobian_fn(single_y))[1]
    for logdet in batched_logdet:
      np.testing.assert_allclose(logdet, logdet_numerical, atol=5e-5)

  def test_raises_on_invalid_input_shape(self):
    bij = LowerUpperTriangularAffine(matrix=jnp.eye(4), bias=jnp.zeros((4,)))
    for fn in [bij.forward, bij.inverse,
               bij.forward_log_det_jacobian, bij.inverse_log_det_jacobian,
               bij.forward_and_log_det, bij.inverse_and_log_det]:
      with self.assertRaises(ValueError):
        fn(jnp.array(0))

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = LowerUpperTriangularAffine(
        matrix=jnp.eye(4),
        bias=jnp.zeros((4,)))
    x = np.zeros((4,))
    f(x, bijector)

  def test_same_as_itself(self):
    bij = LowerUpperTriangularAffine(matrix=jnp.eye(4), bias=jnp.zeros((4,)))
    self.assertTrue(bij.same_as(bij))
    self.assertTrue(bij.same_as(copy.copy(bij)))

  def test_not_same_as_others(self):
    bij = LowerUpperTriangularAffine(matrix=jnp.eye(4), bias=jnp.zeros((4,)))
    other = LowerUpperTriangularAffine(matrix=jnp.eye(4), bias=jnp.ones((4,)))
    self.assertFalse(bij.same_as(other))
    self.assertFalse(bij.same_as(Tanh()))


if __name__ == '__main__':
  jax.config.update('jax_threefry_partitionable', False)
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Masked coupling bijector."""

from typing import Any, Callable, Optional, Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.utils import conversion
from distrax._src.utils import math
import jax.numpy as jnp


Array = base.Array
BijectorParams = Any


class MaskedCoupling(base.Bijector):
  """Coupling bijector that uses a mask to specify which inputs are transformed.

  This coupling bijector takes in a boolean mask that indicates which inputs are
  transformed. Inputs where `mask==True` remain unchanged. Inputs where
  `mask==False` are transformed by an inner bijector, conditioned on the masked
  inputs.

  The number of event dimensions this bijector operates on is referred to as
  `event_ndims`, and is equal to both `event_ndims_in` and `event_ndims_out`.
  By default, `event_ndims` is equal to `mask.ndim + inner_event_ndims`. The
  user can override this by passing an explicit value for `event_ndims`. If
  `event_ndims > mask.ndim + inner_event_ndims`, the mask is broadcast to the
  extra dimensions. If `event_ndims < mask.ndims + inner_event_ndims`, the mask
  is assumed to be a batch of masks that will broadcast against the input.

  Let `f` be a conditional bijector (the inner bijector), `g` be a function (the
  conditioner), and `m` be a boolean mask interpreted numerically, such that
  True is 1 and False is 0. The masked coupling bijector is defined as follows:

  - Forward: `y = (1-m) * f(x; g(m*x)) + m*x`

  - Forward Jacobian log determinant:
    `log|det J(x)| = sum((1-m) * log|df/dx(x; g(m*x))|)`

  - Inverse: `x = (1-m) * f^{-1}(y; g(m*y)) + m*y`

  - Inverse Jacobian log determinant:
    `log|det J(y)| = sum((1-m) * log|df^{-1}/dy(y; g(m*y))|)`
  """

  def __init__(self,
               mask: Array,
               conditioner: Callable[[Array], BijectorParams],
               bijector: Callable[[BijectorParams], base.BijectorLike],
               event_ndims: Optional[int] = None,
               inner_event_ndims: int = 0):
    """Initializes a MaskedCoupling bijector.

    Args:
      mask: the mask, or a batch of masks. Its elements must be boolean; a value
        of True indicates that the corresponding input remains unchanged, and a
        value of False indicates that the corresponding input is transformed.
        The mask should have `mask.ndim` equal to the number of batch dimensions
        plus `event_ndims - inner_event_ndims`. In particular, an inner event
        is either fully masked or fully un-masked: it is not possible to be
        partially masked.
      conditioner: a function that computes the parameters of the inner bijector
        as a function of the masked input. The output of the conditioner will be
        passed to `bijector` in order to obtain the inner bijector.
      bijector: a callable that returns the inner bijector that will be used to
        transform the input. The input to `bijector` is a set of parameters that
        can be used to configure the inner bijector. The `event_ndims_in` and
        `event_ndims_out` of this bijector must match the `inner_event_dims`.
        For example, if `inner_event_dims` is `0`, then the inner bijector must
        be a scalar bijector.
      event_ndims: the number of array dimensions the bijector operates on. If
        None, it defaults to `mask.ndim + inner_event_ndims`. Both
        `event_ndims_in` and `event_ndims_out` are equal to `event_ndims`. Note
        that `event_ndims` should be at least as large as `inner_event_ndims`.
      inner_event_ndims: the number of array dimensions the inner bijector
        operates on. This is `0` by default, meaning the inner bijector acts on
        scalars.
    """
    if mask.dtype != bool:
      raise ValueError(f'`mask` must have values of type `bool`; got values of'
                       f' type `{mask.dtype}`.')
    if event_ndims is not None and event_ndims < inner_event_ndims:
      raise ValueError(f'`event_ndims={event_ndims}` should be at least as'
                       f' large as `inner_event_ndims={inner_event_ndims}`.')
    self._mask = mask
    self._event_mask = jnp.reshape(mask, mask.shape + (1,) * inner_event_ndims)
    self._conditioner = conditioner
    self._bijector = bijector
    self._inner_event_ndims = inner_event_ndims
    if event_ndims is None:
      self._event_ndims = mask.ndim + inner_event_ndims
    else:
      self._event_ndims = event_ndims
    super().__init__(event_ndims_in=self._event_ndims)

  @property
  def bijector(self) -> Callable[[BijectorParams], base.BijectorLike]:
    """The callable that returns the inner bijector of `MaskedCoupling`."""
    return self._bijector

  @property
  def conditioner(self) -> Callable[[Array], BijectorParams]:
    """The conditioner function."""
    return self._conditioner

  @property
  def mask(self) -> Array:
    """The mask characterizing the `MaskedCoupling`, with boolean `dtype`."""
    return self._mask

  def _inner_bijector(self, params: BijectorParams) -> base.Bijector:
    bijector = conversion.as_bijector(self._bijector(params))
    if (bijector.event_ndims_in != self._inner_event_ndims
        or bijector.event_ndims_out != self._inner_event_ndims):
      raise ValueError(
          'The inner bijector event ndims in and out must match the'
          f' `inner_event_ndims={self._inner_event_ndims}`. Instead, got'
          f' `event_ndims_in={bijector.event_ndims_in}` and'
          f' `event_ndims_out={bijector.event_ndims_out}`.')
    return bijector

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    self._check_forward_input_shape(x)
    masked_x = jnp.where(self._event_mask, x, 0.)
    params = self._conditioner(masked_x)
    y0, log_d = self._inner_bijector(params).forward_and_log_det(x)
    y = jnp.where(self._event_mask, x, y0)
    logdet = math.sum_last(
        jnp.where(self._mask, 0., log_d),
        self._event_ndims - self._inner_event_ndims)
    return y, logdet

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    self._check_inverse_input_shape(y)
    masked_y = jnp.where(self._event_mask, y, 0.)
    params = self._conditioner(masked_y)
    x0, log_d = self._inner_bijector(params).inverse_and_log_det(y)
    x = jnp.where(self._event_mask, y, x0)
    logdet = math.sum_last(jnp.where(self._mask, 0., log_d),
                           self._event_ndims - self._inner_event_ndims)
    return x, logdet



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Monte-Carlo estimation of the KL divergence."""

from typing import Optional

import chex
from distrax._src.distributions.distribution import DistributionLike
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions
PRNGKey = chex.PRNGKey


def estimate_kl_best_effort(
    distribution_a: DistributionLike,
    distribution_b: DistributionLike,
    rng_key: PRNGKey,
    num_samples: int,
    proposal_distribution: Optional[DistributionLike] = None):
  """Estimates KL(distribution_a, distribution_b) exactly or with DiCE.

  If the kl_divergence(distribution_a, distribution_b) is not supported,
  the DiCE estimator is used instead.

  Args:
    distribution_a: The first distribution.
    distribution_b: The second distribution.
    rng_key: The PRNGKey random key.
    num_samples: The number of samples, if using the DiCE estimator.
    proposal_distribution: A proposal distribution for the samples, if using
      the DiCE estimator. If None, use `distribution_a` as proposal.

  Returns:
    The estimated KL divergence.
  """
  distribution_a = conversion.as_distribution(distribution_a)
  distribution_b = conversion.as_distribution(distribution_b)
  # If possible, compute the exact KL.
  try:
    return tfd.kl_divergence(distribution_a, distribution_b)
  except NotImplementedError:
    pass
  return mc_estimate_kl(distribution_a, distribution_b, rng_key,
                        num_samples=num_samples,
                        proposal_distribution=proposal_distribution)


def mc_estimate_kl(
    distribution_a: DistributionLike,
    distribution_b: DistributionLike,
    rng_key: PRNGKey,
    num_samples: int,
    proposal_distribution: Optional[DistributionLike] = None):
  """Estimates KL(distribution_a, distribution_b) with the DiCE estimator.

  To get correct gradients with respect the `distribution_a`, we use the DiCE
  estimator, i.e., we stop the gradient with respect to the samples and with
  respect to the denominator in the importance weights. We then do not need
  reparametrized distributions.

  Args:
    distribution_a: The first distribution.
    distribution_b: The second distribution.
    rng_key: The PRNGKey random key.
    num_samples: The number of samples, if using the DiCE estimator.
    proposal_distribution: A proposal distribution for the samples, if using the
      DiCE estimator. If None, use `distribution_a` as proposal.

  Returns:
    The estimated KL divergence.
  """
  if proposal_distribution is None:
    proposal_distribution = distribution_a
  proposal_distribution = conversion.as_distribution(proposal_distribution)
  distribution_a = conversion.as_distribution(distribution_a)
  distribution_b = conversion.as_distribution(distribution_b)

  samples, logp_proposal = proposal_distribution.sample_and_log_prob(
      seed=rng_key, sample_shape=[num_samples])
  samples = jax.lax.stop_gradient(samples)
  logp_proposal = jax.lax.stop_gradient(logp_proposal)
  logp_a = distribution_a.log_prob(samples)
  logp_b = distribution_b.log_prob(samples)
  importance_weight = jnp.exp(logp_a - logp_proposal)
  log_ratio = logp_b - logp_a
  kl_estimator = -importance_weight * log_ratio
  return jnp.mean(kl_estimator, axis=0)


def mc_estimate_kl_with_reparameterized(
    distribution_a: DistributionLike,
    distribution_b: DistributionLike,
    rng_key: PRNGKey,
    num_samples: int):
  """Estimates KL(distribution_a, distribution_b)."""
  if isinstance(distribution_a, tfd.Distribution):
    if distribution_a.reparameterization_type != tfd.FULLY_REPARAMETERIZED:
      raise ValueError(
          f'Distribution `{distribution_a.name}` cannot be reparameterized.')
  distribution_a = conversion.as_distribution(distribution_a)
  distribution_b = conversion.as_distribution(distribution_b)

  samples, logp_a = distribution_a.sample_and_log_prob(
      seed=rng_key, sample_shape=[num_samples])
  logp_b = distribution_b.log_prob(samples)
  log_ratio = logp_b - logp_a
  kl_estimator = -log_ratio
  return jnp.mean(kl_estimator, axis=0)


def mc_estimate_mode(
    distribution: DistributionLike,
    rng_key: PRNGKey,
    num_samples: int):
  """Returns a Monte Carlo estimate of the mode of a distribution."""
  distribution = conversion.as_distribution(distribution)
  # Obtain samples from the distribution and their log probability.
  samples, log_probs = distribution.sample_and_log_prob(
      seed=rng_key, sample_shape=[num_samples])
  # Do argmax over the sample_shape.
  index = jnp.expand_dims(jnp.argmax(log_probs, axis=0), axis=0)
  # Broadcast index to include event_shape of the sample.
  index = index.reshape(index.shape + (1,) * (samples.ndim - index.ndim))
  mode = jnp.squeeze(jnp.take_along_axis(samples, index, axis=0), axis=0)
  return mode



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Distrax bijector for automatically turning JAX functions into Bijectors."""

from typing import Callable, Optional, Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.utils import transformations

Array = base.Array


class Lambda(base.Bijector):
  """Wrapper to automatically turn JAX functions into fully fledged bijectors.

  This class takes in JAX functions that implement bijector methods (such as
  `forward`, `inverse`, `forward_log_det_jacobian`, etc.), and constructs a
  bijector out of them. Any functions not explicitly specified by the user will
  be automatically derived from the existing functions where possible, by
  tracing their JAXPR representation. Missing functions will be derived on
  demand: if a missing function is not used, it will not be derived. At a
  minimum, either `forward` or `inverse` must be given; all other methods will
  be derived (where possible).

  The Lambda bijector can be useful for creating simple one-line bijectors that
  would otherwise be tedious to define. Examples of scalar bijectors that can be
  easily constructed with Lambda are:

  - Identity: `Lambda(lambda x: x)`
  - Affine: `Lambda(lambda x: a*x + b)`
  - Tanh: `Lambda(jnp.tanh)`
  - Composite: `Lambda(lambda x: jnp.tanh(a*x + b))`

  Requirements and limitations:

  - Only functions composed entirely of invertible primitives can be
    automatically inverted (see `bijection_utils.py` for a list of invertible
    primitives). If the inverse is needed but is not automatically derivable,
    the user must provide it explicitly.

  - If log-determinant functions are not provided, Lambda will assume that
    `forward` and `inverse` are scalar functions applied elementwise. If the
    bijector is not meant to be scalar, its log-determinant functions must be
    provided explicitly by the user.
  """

  def __init__(
      self,
      forward: Optional[Callable[[Array], Array]] = None,
      inverse: Optional[Callable[[Array], Array]] = None,
      forward_log_det_jacobian: Optional[Callable[[Array], Array]] = None,
      inverse_log_det_jacobian: Optional[Callable[[Array], Array]] = None,
      event_ndims_in: Optional[int] = None,
      event_ndims_out: Optional[int] = None,
      is_constant_jacobian: Optional[bool] = None):
    """Initializes a Lambda bijector with methods specified as args."""

    if forward is None and inverse is None:
      raise ValueError("The Lambda bijector requires at least one of `forward` "
                       "or `inverse` to be specified, but neither is.")

    jac_functions_specified = (forward_log_det_jacobian is not None
                               or inverse_log_det_jacobian is not None)
    if jac_functions_specified:
      if event_ndims_in is None:
        raise ValueError("When log det Jacobian functions are specified, you "
                         "must also specify `event_ndims_in`.")
    else:
      if event_ndims_in is not None or event_ndims_out is not None:
        raise ValueError("When log det Jacobian functions are unspecified, you "
                         "must leave `event_ndims_in` and `event_ndims_out` "
                         "unspecified; they will default to 0.")
      event_ndims_in = 0

    if is_constant_jacobian is None:
      fn = inverse if forward is None else forward
      is_constant_jacobian = transformations.is_constant_jacobian(fn)

    super().__init__(
        event_ndims_in=event_ndims_in,
        event_ndims_out=event_ndims_out,
        is_constant_jacobian=is_constant_jacobian)
    self._forward = forward
    self._inverse = inverse
    self._forward_log_det_jacobian = forward_log_det_jacobian
    self._inverse_log_det_jacobian = inverse_log_det_jacobian

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    self._check_forward_input_shape(x)
    if self._forward is None:
      self._forward = transformations.inv(self._inverse)
    return self._forward(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    self._check_inverse_input_shape(y)
    if self._inverse is None:
      self._inverse = transformations.inv(self._forward)
    return self._inverse(y)

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    self._check_forward_input_shape(x)
    if self._forward_log_det_jacobian is None:
      self._forward_log_det_jacobian = transformations.log_det_scalar(
          self.forward)
    return self._forward_log_det_jacobian(x)

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    self._check_inverse_input_shape(y)
    if self._inverse_log_det_jacobian is None:
      self._inverse_log_det_jacobian = transformations.log_det_scalar(
          self.inverse)
    return self._inverse_log_det_jacobian(y)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self.forward(x), self.forward_log_det_jacobian(x)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    return self.inverse(y), self.inverse_log_det_jacobian(y)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is Lambda:  # pylint: disable=unidiomatic-typecheck
      return all((
          self.forward is other.forward,
          self.inverse is other.inverse,
          self.forward_log_det_jacobian is other.forward_log_det_jacobian,
          self.inverse_log_det_jacobian is other.inverse_log_det_jacobian,
          self.forward_and_log_det is other.forward_and_log_det,
          self.inverse_and_log_det is other.inverse_and_log_det,
      ))

    return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `independent.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import independent
from distrax._src.distributions import mvn_diag
from distrax._src.distributions import normal
from distrax._src.utils import equivalence
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions


class IndependentTest(parameterized.TestCase):
  """Class to test miscellaneous methods of the `Independent` distribution."""

  def setUp(self):
    super().setUp()
    self.loc = np.random.randn(2, 3, 4)
    self.scale = np.abs(np.random.randn(2, 3, 4))
    self.base = normal.Normal(loc=self.loc, scale=self.scale)
    self.dist = independent.Independent(self.base, reinterpreted_batch_ndims=1)

  def assertion_fn(self, rtol):
    return lambda x, y: np.testing.assert_allclose(x, y, rtol=rtol)

  @parameterized.parameters(None, 0, 1, 2)
  def test_constructor_is_jittable_given_ndims(self, ndims):
    constructor = lambda d: independent.Independent(d, ndims)
    jax.jit(constructor)(self.base)  # pylint: disable=not-callable

  @parameterized.named_parameters(
      ('single element', 1, (3,)),
      ('range', slice(-1), (1, 3)),
      ('range_2', (slice(None), slice(-1)), (2, 2)),
  )
  def test_slice(self, slice_, expected_batch_shape):
    sliced_dist = self.dist[slice_]
    self.assertEqual(sliced_dist.batch_shape, expected_batch_shape)
    self.assertEqual(sliced_dist.event_shape, self.dist.event_shape)
    self.assertIsInstance(sliced_dist, independent.Independent)
    self.assertIsInstance(sliced_dist.distribution, self.base.__class__)
    self.assertion_fn(rtol=1e-3)(
        sliced_dist.distribution.loc, self.loc[slice_])
    self.assertion_fn(rtol=1e-3)(
        sliced_dist.distribution.scale, self.scale[slice_])

  def test_slice_ellipsis(self):
    sliced_dist = self.dist[..., -1]
    expected_batch_shape = (2,)
    self.assertEqual(sliced_dist.batch_shape, expected_batch_shape)
    self.assertEqual(sliced_dist.event_shape, self.dist.event_shape)
    self.assertIsInstance(sliced_dist, independent.Independent)
    self.assertIsInstance(sliced_dist.distribution, self.base.__class__)
    self.assertion_fn(rtol=1e-3)(
        sliced_dist.distribution.loc, self.loc[:, -1, :])
    self.assertion_fn(rtol=1e-3)(
        sliced_dist.distribution.scale, self.scale[:, -1, :])

  def test_vmap_inputs(self):
    def log_prob_sum(dist, x):
      return dist.log_prob(x).sum()

    base = normal.Normal(
        jnp.arange(3 * 4 * 5).reshape((3, 4, 5)), jnp.ones((3, 4, 5)))
    dist = independent.Independent(base, reinterpreted_batch_ndims=1)
    x = jnp.zeros((3, 4, 5))

    with self.subTest('no vmap'):
      actual = log_prob_sum(dist, x)
      expected = dist.log_prob(x).sum()
      self.assertion_fn(rtol=1e-6)(actual, expected)

    with self.subTest('axis=0'):
      actual = jax.vmap(log_prob_sum, in_axes=0)(dist, x)
      expected = dist.log_prob(x).sum(axis=1)
      self.assertion_fn(rtol=1e-6)(actual, expected)

    with self.subTest('axis=1'):
      actual = jax.vmap(log_prob_sum, in_axes=1)(dist, x)
      expected = dist.log_prob(x).sum(axis=0)
      self.assertion_fn(rtol=1e-6)(actual, expected)

  def test_vmap_outputs(self):
    def summed_dist(loc, scale):
      return independent.Independent(
          normal.Normal(loc.sum(keepdims=True), scale.sum(keepdims=True)),
          reinterpreted_batch_ndims=1)

    loc = jnp.arange((3 * 4 * 5)).reshape((3, 4, 5))
    scale = jnp.ones((3, 4, 5))

    actual = jax.vmap(summed_dist, in_axes=0)(loc, scale)
    expected = independent.Independent(
        normal.Normal(loc.sum(axis=(1, 2), keepdims=True),
                      scale.sum(axis=(1, 2), keepdims=True)),
        reinterpreted_batch_ndims=1)

    np.testing.assert_equal(actual.batch_shape, expected.batch_shape)
    np.testing.assert_equal(actual.event_shape, expected.event_shape)

    x = jnp.array([[[1]], [[2]], [[3]]])
    self.assertion_fn(rtol=1e-6)(actual.log_prob(x),
                                 expected.log_prob(x))


class TFPMultivariateNormalTest(equivalence.EquivalenceTest):
  """Class to test Distrax Independent distribution against its TFP counterpart.

  This class tests the case when using a TFP multivariate Normal distribution
  as input for the TFP and Distrax Independent. There are 2 methods to create
  the base distributions, `_make_base_distribution` and
  `_make_tfp_base_distribution`. By overloading these methods, different
  base distributions can be used.
  """

  def _make_tfp_base_distribution(self, loc, scale):
    return tfd.MultivariateNormalDiag(loc=loc, scale_diag=scale)

  # Define the function to create the base distribution.
  _make_base_distribution = _make_tfp_base_distribution

  def setUp(self):
    super().setUp()
    self._init_distr_cls(independent.Independent)

    self.normal_loc = jax.random.normal(
        key=jax.random.PRNGKey(42), shape=(5, 4, 3, 2))
    self.normal_scale = 0.5 + np.abs(
        jax.random.normal(key=jax.random.PRNGKey(43), shape=(5, 4, 3, 2)))
    self.normal_loc2 = jax.random.normal(
        key=jax.random.PRNGKey(43), shape=(5, 4, 3, 2))
    self.normal_scale2 = 0.5 + np.abs(
        jax.random.normal(key=jax.random.PRNGKey(44), shape=(5, 4, 3, 2)))

    # For most tests, we use `base_dist` and `tfp_base_dist` as base
    # distributions. The latter is used as input for the TFP Independent, which
    # we compare against.
    self.base_dist = self._make_base_distribution(self.normal_loc,
                                                  self.normal_scale)
    self.tfp_base_dist = self._make_tfp_base_distribution(self.normal_loc,
                                                          self.normal_scale)
    # Some methods (e.g., the KL divergence) require two distributions. We
    # define here the base distribution for those methods.
    self.base_dist2 = self._make_base_distribution(self.normal_loc2,
                                                   self.normal_scale2)
    self.tfp_base_dist2 = self._make_tfp_base_distribution(self.normal_loc2,
                                                           self.normal_scale2)

  def test_invalid_parameters(self):
    self._test_raises_error(
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': -1})
    self._test_raises_error(
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': 10})

  @parameterized.named_parameters(
      ('batch dims None', None),
      ('batch dims 0', 0),
      ('batch dims 1', 1),
      ('batch dims 2', 2),
      ('batch dims 3', 3),
  )
  def test_event_shape(self, batch_ndims):
    super()._test_event_shape(
        (),
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': batch_ndims},
        tfp_dist_kwargs={'distribution': self.tfp_base_dist,
                         'reinterpreted_batch_ndims': batch_ndims},
    )

  @parameterized.named_parameters(
      ('batch dims None', None),
      ('batch dims 0', 0),
      ('batch dims 1', 1),
      ('batch dims 2', 2),
      ('batch dims 3', 3),
  )
  def test_batch_shape(self, batch_ndims):
    super()._test_batch_shape(
        (),
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': batch_ndims},
        tfp_dist_kwargs={'distribution': self.tfp_base_dist,
                         'reinterpreted_batch_ndims': batch_ndims},
    )

  @chex.all_variants
  @parameterized.named_parameters(
      ('batch dims None, empty shape', None, ()),
      ('batch dims None, int shape', None, 10),
      ('batch dims None, 2-tuple shape', None, (10, 20)),
      ('batch dims 1, empty shape', 1, ()),
      ('batch dims 1, int shape', 1, 10),
      ('batch dims 1, 2-tuple shape', 1, (10, 20)),
      ('batch dims 3, empty shape', 3, ()),
      ('batch dims 3, int shape', 3, 10),
      ('batch dims 3, 2-tuple shape', 3, (10, 20)),
  )
  def test_sample_shape(self, batch_ndims, sample_shape):
    super()._test_sample_shape(
        (),
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': batch_ndims},
        tfp_dist_kwargs={'distribution': self.tfp_base_dist,
                         'reinterpreted_batch_ndims': batch_ndims},
        sample_shape=sample_shape)

  @chex.all_variants
  def test_sample_dtype(self):
    dist = self.distrax_cls(self.base_dist)
    samples = self.variant(dist.sample)(seed=self.key)
    self.assertEqual(dist.dtype, samples.dtype)
    self.assertEqual(dist.dtype, self.base_dist.dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('batch dims None, empty shape', None, ()),
      ('batch dims None, int shape', None, 10),
      ('batch dims None, 2-tuple shape', None, (10, 20)),
      ('batch dims 1, empty shape', 1, ()),
      ('batch dims 1, int shape', 1, 10),
      ('batch dims 1, 2-tuple shape', 1, (10, 20)),
      ('batch dims 3, empty shape', 3, ()),
      ('batch dims 3, int shape', 3, 10),
      ('batch dims 3, 2-tuple shape', 3, (10, 20)),
  )
  def test_sample_and_log_prob(self, batch_ndims, sample_shape):
    super()._test_sample_and_log_prob(
        dist_args=(),
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': batch_ndims},
        tfp_dist_kwargs={'distribution': self.tfp_base_dist,
                         'reinterpreted_batch_ndims': batch_ndims},
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('batch dims None', None, np.zeros((5, 4, 3, 2))),
      ('batch dims 0', 0, np.zeros((5, 4, 3, 2))),
      ('batch dims 1', 1, np.zeros((5, 4, 3, 2))),
      ('batch dims 2', 2, np.zeros((5, 4, 3, 2))),
      ('batch dims 3', 3, np.zeros((5, 4, 3, 2))),
  )
  def test_log_prob(self, batch_ndims, value):
    super()._test_attribute(
        attribute_string='log_prob',
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': batch_ndims},
        tfp_dist_kwargs={'distribution': self.tfp_base_dist,
                         'reinterpreted_batch_ndims': batch_ndims},
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('entropy, batch dims None', 'entropy', None),
      ('entropy, batch dims 1', 'entropy', 1),
      ('entropy, batch dims 3', 'entropy', 3),
      ('mean, batch dims None', 'mean', None),
      ('mean, batch dims 1', 'mean', 1),
      ('mean, batch dims 3', 'mean', 3),
      ('variance, batch dims None', 'variance', None),
      ('variance, batch dims 1', 'variance', 1),
      ('variance, batch dims 3', 'variance', 3),
      ('stddev, batch dims None', 'stddev', None),
      ('stddev, batch dims 1', 'stddev', 1),
      ('stddev, batch dims 3', 'stddev', 3),
      ('mode, batch dims None', 'mode', None),
      ('mode, batch dims 1', 'mode', 1),
      ('mode, batch dims 3', 'mode', 3),
  )
  def test_method(self, function_string, batch_ndims):
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': batch_ndims},
        tfp_dist_kwargs={'distribution': self.tfp_base_dist,
                         'reinterpreted_batch_ndims': batch_ndims},
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants(with_jit=False, with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax, no batch dims',
       'kl_divergence', 'distrax_to_distrax', None),
      ('kl distrax_to_tfp, no batch dims',
       'kl_divergence', 'distrax_to_tfp', None),
      ('kl tfp_to_distrax, no batch dims',
       'kl_divergence', 'tfp_to_distrax', None),
      ('cross-ent distrax_to_distrax, no batch dims',
       'cross_entropy', 'distrax_to_distrax', None),
      ('cross-ent distrax_to_tfp, no batch dims',
       'cross_entropy', 'distrax_to_tfp', None),
      ('cross-ent tfp_to_distrax, no batch dims',
       'cross_entropy', 'tfp_to_distrax', None),
      ('kl distrax_to_distrax, batch dims 2',
       'kl_divergence', 'distrax_to_distrax', 2),
      ('kl distrax_to_tfp, batch dims 2',
       'kl_divergence', 'distrax_to_tfp', 2),
      ('kl tfp_to_distrax, batch dims 2',
       'kl_divergence', 'tfp_to_distrax', 2),
      ('cross-ent distrax_to_distrax, batch dims 2',
       'cross_entropy', 'distrax_to_distrax', 2),
      ('cross-ent distrax_to_tfp, batch dims 2',
       'cross_entropy', 'distrax_to_tfp', 2),
      ('cross-ent tfp_to_distrax, batch dims 2',
       'cross_entropy', 'tfp_to_distrax', 2),
  )
  def test_with_two_distributions(self, function_string, mode_string,
                                  batch_ndims):
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={'distribution': self.base_dist,
                      'reinterpreted_batch_ndims': batch_ndims},
        dist2_kwargs={'distribution': self.base_dist2,
                      'reinterpreted_batch_ndims': batch_ndims},
        tfp_dist1_kwargs={'distribution': self.tfp_base_dist,
                          'reinterpreted_batch_ndims': batch_ndims},
        tfp_dist2_kwargs={'distribution': self.tfp_base_dist2,
                          'reinterpreted_batch_ndims': batch_ndims},
        assertion_fn=self.assertion_fn(rtol=1e-3))


class TFPUnivariateNormalTest(TFPMultivariateNormalTest):
  """Class to test Distrax Independent distribution against its TFP counterpart.

  This class tests the case when using a TFP univariate Normal distribution
  as input for the TFP and distrax Independent.
  """

  def _make_tfp_base_distribution(self, loc, scale):
    return tfd.Normal(loc=loc, scale=scale)

  # Define the function to create the base distribution.
  _make_base_distribution = _make_tfp_base_distribution

  def test_jittable(self):
    super()._test_jittable(
        dist_kwargs={'distribution': self.base_dist,
                     'reinterpreted_batch_ndims': 1},
        assertion_fn=self.assertion_fn(rtol=1e-4))


class DistraxUnivariateNormalTest(TFPMultivariateNormalTest):
  """Class to test Distrax Independent distribution against its TFP counterpart.

  This class tests the case when using a distrax univariate Normal distribution
  as input for the Distrax Independent.
  """

  def _make_distrax_base_distribution(self, loc, scale):
    return normal.Normal(loc=loc, scale=scale)

  def _make_tfp_base_distribution(self, loc, scale):
    return tfd.Normal(loc=loc, scale=scale)

  # Define the function to create the base distribution.
  _make_base_distribution = _make_distrax_base_distribution


class DistraxMultivariateNormalTest(TFPMultivariateNormalTest):
  """Class to test Distrax Independent distribution against its TFP counterpart.

  This class tests the case when using a Distrax multivariate Normal
  distribution as input for the Distrax Independent.
  """

  def _make_distrax_base_distribution(self, loc, scale):
    return mvn_diag.MultivariateNormalDiag(loc=loc, scale_diag=scale)

  # Define the function to create the base distribution.
  _make_base_distribution = _make_distrax_base_distribution


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `lambda_bijector.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import lambda_bijector
from distrax._src.distributions import normal
from distrax._src.distributions import transformed
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions
tfb = tfp.bijectors


RTOL = 1e-2


def _with_additional_parameters(params, all_named_parameters):
  """Convenience function for appending a cartesian product of parameters."""
  for name, param in params:
    for named_params in all_named_parameters:
      yield (f'{named_params[0]}; {name}',) + named_params[1:] + (param,)


def _with_base_dists(*all_named_parameters):
  """Partial of _with_additional_parameters to specify distrax and TFP base."""
  base_dists = (
      ('tfp_base', tfd.Normal),
      ('distrax_base', normal.Normal),
  )
  return _with_additional_parameters(base_dists, all_named_parameters)


class LambdaTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.seed = jax.random.PRNGKey(1234)

  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal', 0, 1),
      ('2d std normal', np.zeros(2), np.ones(2)),
      ('broadcasted loc', 0, np.ones(3)),
      ('broadcasted scale', np.ones(3), 1),
  ))
  def test_event_shape(self, mu, sigma, base_dist):
    base = base_dist(mu, sigma)
    bijector = lambda_bijector.Lambda(jnp.tanh)
    dist = transformed.Transformed(base, bijector)

    tfp_bijector = tfb.Tanh()
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)

    assert dist.event_shape == tfp_dist.event_shape

  def test_raises_on_both_none(self):
    with self.assertRaises(ValueError):
      lambda_bijector.Lambda(forward=None, inverse=None)

  def test_raises_on_log_det_without_event_ndims(self):
    with self.assertRaises(ValueError):
      lambda_bijector.Lambda(
          forward=lambda x: x,
          forward_log_det_jacobian=lambda x: jnp.zeros_like(x[:-1]),
          event_ndims_in=None)

  @parameterized.named_parameters(
      ('event_ndims_in', 0, None),
      ('event_ndims_out', None, 0),
      ('event_ndims_in and event_ndims_out', 0, 0),
  )
  def test_raises_on_event_ndims_without_log_det(self, ndims_in, ndims_out):
    with self.assertRaises(ValueError):
      lambda_bijector.Lambda(
          forward=lambda x: x,
          event_ndims_in=ndims_in,
          event_ndims_out=ndims_out)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal, no shape', 0, 1, ()),
      ('1d std normal, int shape', 0, 1, 1),
      ('1d std normal, 1-tuple shape', 0, 1, (1,)),
      ('1d std normal, 2-tuple shape', 0, 1, (2, 2)),
      ('2d std normal, no shape', np.zeros(2), np.ones(2), ()),
      ('2d std normal, int shape', [0, 0], [1, 1], 1),
      ('2d std normal, 1-tuple shape', np.zeros(2), np.ones(2), (1,)),
      ('2d std normal, 2-tuple shape', [0, 0], [1, 1], (2, 2)),
      ('rank 2 std normal, 2-tuple shape', np.zeros(
          (3, 2)), np.ones((3, 2)), (2, 2)),
      ('broadcasted loc', 0, np.ones(3), (2, 2)),
      ('broadcasted scale', np.ones(3), 1, ()),
  ))
  def test_sample_shape(self, mu, sigma, sample_shape, base_dist):
    base = base_dist(mu, sigma)
    bijector = lambda_bijector.Lambda(jnp.tanh)
    dist = transformed.Transformed(base, bijector)
    def sample_fn(seed, sample_shape):
      return dist.sample(seed=seed, sample_shape=sample_shape)
    samples = self.variant(sample_fn, ignore_argnums=(1,), static_argnums=1)(
        self.seed, sample_shape)

    tfp_bijector = tfb.Tanh()
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    tfp_samples = tfp_dist.sample(sample_shape=sample_shape,
                                  seed=self.seed)

    chex.assert_equal_shape([samples, tfp_samples])

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d dist, 1d value', 0, 1, 0.5),
      ('1d dist, 2d value', 0., 1., np.array([0.25, 0.5])),
      ('2d dist, 1d value', np.zeros(2), np.ones(2), 0.5),
      ('2d broadcasted dist, 1d value', np.zeros(2), 1, 0.5),
      ('2d dist, 2d value', np.zeros(2), np.ones(2), np.array([0.25, 0.5])),
      ('1d dist, 1d value, edge case', 0, 1, 0.99),
  ))
  def test_log_prob(self, mu, sigma, value, base_dist):
    base = base_dist(mu, sigma)
    bijector = lambda_bijector.Lambda(jnp.tanh)
    dist = transformed.Transformed(base, bijector)
    actual = self.variant(dist.log_prob)(value)

    tfp_bijector = tfb.Tanh()
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    expected = tfp_dist.log_prob(value)

    np.testing.assert_allclose(actual, expected, rtol=RTOL)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d dist, 1d value', 0, 1, 0.5),
      ('1d dist, 2d value', 0., 1., np.array([0.25, 0.5])),
      ('2d dist, 1d value', np.zeros(2), np.ones(2), 0.5),
      ('2d broadcasted dist, 1d value', np.zeros(2), 1, 0.5),
      ('2d dist, 2d value', np.zeros(2), np.ones(2), np.array([0.25, 0.5])),
      ('1d dist, 1d value, edge case', 0, 1, 0.99),
  ))
  def test_prob(self, mu, sigma, value, base_dist):
    base = base_dist(mu, sigma)
    bijector = lambda_bijector.Lambda(jnp.tanh)
    dist = transformed.Transformed(base, bijector)
    actual = self.variant(dist.prob)(value)

    tfp_bijector = tfb.Tanh()
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    expected = tfp_dist.prob(value)

    np.testing.assert_allclose(actual, expected, rtol=RTOL)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal, no shape', 0, 1, ()),
      ('1d std normal, int shape', 0, 1, 1),
      ('1d std normal, 1-tuple shape', 0, 1, (1,)),
      ('1d std normal, 2-tuple shape', 0, 1, (2, 2)),
      ('2d std normal, no shape', np.zeros(2), np.ones(2), ()),
      ('2d std normal, int shape', [0, 0], [1, 1], 1),
      ('2d std normal, 1-tuple shape', np.zeros(2), np.ones(2), (1,)),
      ('2d std normal, 2-tuple shape', [0, 0], [1, 1], (2, 2)),
      ('rank 2 std normal, 2-tuple shape', np.zeros(
          (3, 2)), np.ones((3, 2)), (2, 2)),
      ('broadcasted loc', 0, np.ones(3), (2, 2)),
      ('broadcasted scale', np.ones(3), 1, ()),
  ))
  def test_sample_and_log_prob(self, mu, sigma, sample_shape, base_dist):
    base = base_dist(mu, sigma)
    bijector = lambda_bijector.Lambda(lambda x: 10 * jnp.tanh(0.1 * x))
    dist = transformed.Transformed(base, bijector)
    def sample_and_log_prob_fn(seed, sample_shape):
      return dist.sample_and_log_prob(seed=seed, sample_shape=sample_shape)
    samples, log_prob = self.variant(
        sample_and_log_prob_fn, ignore_argnums=(1,), static_argnums=(1,))(
            self.seed, sample_shape)
    expected_samples = bijector.forward(
        base.sample(seed=self.seed, sample_shape=sample_shape))

    tfp_bijector = tfb.Chain([tfb.Scale(10), tfb.Tanh(), tfb.Scale(0.1)])
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    tfp_samples = tfp_dist.sample(seed=self.seed, sample_shape=sample_shape)
    tfp_log_prob = tfp_dist.log_prob(samples)

    chex.assert_equal_shape([samples, tfp_samples])
    np.testing.assert_allclose(log_prob, tfp_log_prob, rtol=RTOL)
    np.testing.assert_allclose(samples, expected_samples, rtol=RTOL)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(_with_base_dists(
      ('entropy', 'entropy', 0., 1.),
      ('mean', 'mean', 0, 1),
      ('mean from list params', 'mean', [-1, 1], [1, 2]),
      ('mode', 'mode', 0, 1),
  ))
  def test_method(self, function_string, mu, sigma, base_dist):
    base = base_dist(mu, sigma)
    bijector = lambda_bijector.Lambda(lambda x: x + 3)
    dist = transformed.Transformed(base, bijector)

    tfp_bijector = tfb.Shift(3)
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)

    np.testing.assert_allclose(self.variant(getattr(dist, function_string))(),
                               getattr(tfp_dist, function_string)(), rtol=RTOL)

  @chex.all_variants(with_jit=False)  # no need to jit function transformations
  @parameterized.named_parameters(
      ('identity', lambda x: x, tfb.Identity),
      ('tanh', jnp.tanh, tfb.Tanh),
      ('scale', lambda x: 3.0 * x, lambda: tfb.Scale(3.0)),
      ('shift', lambda x: x + 2.0, lambda: tfb.Shift(2.0)),
      ('exp', jnp.exp, tfb.Exp),
      ('softplus', lambda x: jnp.log1p(jnp.exp(x)), tfb.Softplus),
      ('square', jnp.square, tfb.Square),
  )
  def test_log_dets(self, lambda_bjct, tfp_bijector_fn):
    bijector = lambda_bijector.Lambda(lambda_bjct)
    tfp_bijector = tfp_bijector_fn()

    x = np.array([0.05, 0.3, 0.45], dtype=np.float32)
    fldj = tfp_bijector.forward_log_det_jacobian(x, event_ndims=0)
    fldj_ = self.variant(bijector.forward_log_det_jacobian)(x)
    np.testing.assert_allclose(fldj_, fldj, rtol=RTOL)

    y = bijector.forward(x)  # pytype: disable=wrong-arg-types  # jax-ndarray
    ildj = tfp_bijector.inverse_log_det_jacobian(y, event_ndims=0)
    ildj_ = self.variant(bijector.inverse_log_det_jacobian)(y)
    np.testing.assert_allclose(ildj_, ildj, rtol=RTOL)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('identity', lambda x: x, tfb.Identity),
      ('tanh', jnp.tanh, tfb.Tanh),
      ('scale', lambda x: 3.0 * x, lambda: tfb.Scale(3.0)),
      ('shift', lambda x: x + 2.0, lambda: tfb.Shift(2.0)),
      ('exp', jnp.exp, tfb.Exp),
      ('softplus', lambda x: jnp.log1p(jnp.exp(x)), tfb.Softplus),
      ('square', jnp.square, tfb.Square),
  ))
  def test_against_tfp_bijectors(
      self, lambda_bjct, tfp_bijector, base_dist):
    mu = np.array([-1.0, 0.0, 1.0], dtype=np.float32)
    sigma = np.array([0.5, 1.0, 2.5], dtype=np.float32)
    base = base_dist(mu, sigma)

    bijector = lambda_bijector.Lambda(lambda_bjct)
    dist = transformed.Transformed(base, bijector)
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector())

    y = np.array([0.05, 0.3, 0.95], dtype=np.float32)

    lp_y = tfp_dist.log_prob(y)
    lp_y_ = self.variant(dist.log_prob)(y)
    np.testing.assert_allclose(lp_y_, lp_y, rtol=RTOL)

    p_y = tfp_dist.prob(y)
    p_y_ = self.variant(dist.prob)(y)
    np.testing.assert_allclose(p_y_, p_y, rtol=RTOL)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('identity', lambda x: x, tfb.Identity),
      ('tanh', jnp.tanh, tfb.Tanh),
      ('scale', lambda x: 3.0 * x, lambda: tfb.Scale(3.0)),
  ))
  def test_auto_lambda(
      self, forward_fn, tfp_bijector, base_dist):
    mu = np.array([-1.0, 0.0, 1.0], dtype=np.float32)
    sigma = np.array([0.5, 1.0, 2.5], dtype=np.float32)
    base = base_dist(mu, sigma)

    dist = transformed.Transformed(base, forward_fn)
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector())

    y = np.array([0.05, 0.3, 0.95], dtype=np.float32)

    lp_y = tfp_dist.log_prob(y)
    lp_y_ = self.variant(dist.log_prob)(y)
    np.testing.assert_allclose(lp_y_, lp_y, rtol=RTOL)

    p_y = tfp_dist.prob(y)
    p_y_ = self.variant(dist.prob)(y)
    np.testing.assert_allclose(p_y_, p_y, rtol=RTOL)

  def test_raises_on_invalid_input_shape(self):
    bij = lambda_bijector.Lambda(
        forward=lambda x: x,
        inverse=lambda y: y,
        forward_log_det_jacobian=lambda x: jnp.zeros_like(x[:-1]),
        inverse_log_det_jacobian=lambda y: jnp.zeros_like(y[:-1]),
        event_ndims_in=1)
    for fn in [bij.forward, bij.inverse,
               bij.forward_log_det_jacobian, bij.inverse_log_det_jacobian,
               bij.forward_and_log_det, bij.inverse_and_log_det]:
      with self.assertRaises(ValueError):
        fn(jnp.array(0))

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = lambda_bijector.Lambda(lambda x: x)
    x = np.zeros(())
    f(x, bijector)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `dirichlet.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import beta
from distrax._src.distributions import dirichlet
from distrax._src.utils import equivalence
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions


class DirichletTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(dirichlet.Dirichlet)

  @parameterized.named_parameters(
      ('1d params', (3,), ()),
      ('2d params', (4, 3), (4,)),
  )
  def test_properties(self, concentration_shape, batch_shape):
    rng = np.random.default_rng(42)
    concentration = 0.1 + rng.uniform(size=concentration_shape)
    dist = self.distrax_cls(concentration)
    self.assertEqual(dist.event_shape, (3,))
    self.assertEqual(dist.batch_shape, batch_shape)
    self.assertion_fn(rtol=1e-4)(dist.concentration, concentration)

  @parameterized.named_parameters(
      ('0d params', ()),
      ('1d params with K=1', (1,)),
      ('2d params with K=1', (4, 1)),
  )
  def test_raises_on_wrong_concentration(self, concentration_shape):
    rng = np.random.default_rng(42)
    concentration = 0.1 + rng.uniform(size=concentration_shape)
    with self.assertRaises(ValueError):
      self.distrax_cls(concentration)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d params, no shape', (3,), ()),
      ('1d params, int shape', (3,), 4),
      ('1d params, 1-tuple shape', (3,), (4,)),
      ('1d params, 2-tuple shape', (3,), (5, 4)),
      ('2d params, no shape', (2, 3), ()),
      ('2d params, int shape', (2, 3), 4),
      ('2d params, 1-tuple shape', (2, 3), (4,)),
      ('2d params, 2-tuple shape', (2, 3), (5, 4)),
  )
  def test_sample_shape(self, concentration_shape, sample_shape):
    rng = np.random.default_rng(42)
    concentration = 0.1 + rng.uniform(size=concentration_shape)
    super()._test_sample_shape((concentration,), dict(), sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float16', jnp.float16))
  def test_sample_dtype(self, dtype):
    dist = self.distrax_cls(concentration=jnp.ones((3,), dtype=dtype))
    samples = self.variant(dist.sample)(seed=self.key)
    self.assertEqual(samples.dtype, dist.dtype)
    self.assertEqual(samples.dtype, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d params, no shape', (3,), ()),
      ('1d params, int shape', (3,), 4),
      ('1d params, 1-tuple shape', (3,), (4,)),
      ('1d params, 2-tuple shape', (3,), (5, 4)),
      ('2d params, no shape', (2, 3), ()),
      ('2d params, int shape', (2, 3), 4),
      ('2d params, 1-tuple shape', (2, 3), (4,)),
      ('2d params, 2-tuple shape', (2, 3), (5, 4)),
  )
  def test_sample_and_log_prob(self, concentration_shape, sample_shape):
    rng = np.random.default_rng(42)
    concentration = 0.1 + rng.uniform(size=concentration_shape)
    super()._test_sample_and_log_prob(
        dist_args=(concentration,),
        dist_kwargs=dict(),
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d params, 1d value', (3,), (3,)),
      ('1d params, 2d value', (3,), (4, 3)),
      ('2d params, 1d value', (4, 3), (3,)),
      ('2d params, 2d value', (4, 3), (4, 3)),
      ('2d params, 3d value', (4, 3), (5, 4, 3)),
  )
  def test_methods_with_value(self, concentration_shape, value_shape):
    rng = np.random.default_rng(42)
    concentration = np.abs(rng.normal(size=concentration_shape))
    value = rng.uniform(size=value_shape)
    value /= np.sum(value, axis=-1, keepdims=True)
    for method in ['prob', 'log_prob']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=(concentration,),
            call_args=(value,),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('1d params', (3,)),
      ('2d params', (4, 3)),
  )
  def test_method(self, concentration_shape):
    rng = np.random.default_rng(42)
    concentration = np.abs(rng.normal(size=concentration_shape))
    for method in ['entropy', 'mean', 'variance', 'stddev', 'covariance']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=(concentration,),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('without nans', (2., 3., 4.), (1. / 6., 2. / 6., 3. / 6.)),
      ('with nans', (0.5, 3., 4.), np.nan),
  )
  def test_mode(self, concentration, expected_result):
    dist = self.distrax_cls(concentration)
    result = self.variant(dist.mode)()
    if np.any(np.isnan(expected_result)):
      self.assertTrue(jnp.all(jnp.isnan(result)))
    else:
      self.assertion_fn(rtol=1e-3)(result, expected_result)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('distrax_to_distrax', 'distrax_to_distrax'),
      ('distrax_to_tfp', 'distrax_to_tfp'),
      ('tfp_to_distrax', 'tfp_to_distrax'),
  )
  def test_with_two_distributions(self, mode_string):
    rng = np.random.default_rng(42)
    concentration1 = np.abs(rng.normal(size=(4, 3, 2))).astype(np.float32)
    concentration2 = np.abs(rng.normal(size=(3, 2))).astype(np.float32)
    for method in ['kl_divergence', 'cross_entropy']:
      with self.subTest(method=method):
        super()._test_with_two_distributions(
            attribute_string=method,
            mode_string=mode_string,
            dist1_kwargs={'concentration': concentration1},
            dist2_kwargs={'concentration': concentration2},
            assertion_fn=self.assertion_fn(rtol=3e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('distrax_to_distrax', 'distrax_to_distrax'),
      ('distrax_to_tfp', 'distrax_to_tfp'),
      ('tfp_to_distrax', 'tfp_to_distrax'),
  )
  def test_with_two_distributions_dirichlet_beta(self, mode_string):
    rng = np.random.default_rng(42)
    # Distribution 1 is Dirichlet.
    alpha1 = rng.uniform(size=(4, 3)).astype(np.float32)
    beta1 = rng.uniform(size=(4, 3)).astype(np.float32)
    concentration1 = np.stack((alpha1, beta1), axis=-1)
    distrax_dist1 = self.distrax_cls(concentration1)
    tfp_dist1 = tfd.Dirichlet(concentration1)
    # Distribution 2 is Beta.
    alpha2 = rng.uniform(size=(3,)).astype(np.float32)
    beta2 = rng.uniform(size=(3,)).astype(np.float32)
    distrax_dist2 = beta.Beta(alpha2, beta2)
    tfp_dist2 = tfd.Beta(alpha2, beta2)
    for method in ['kl_divergence', 'cross_entropy']:
      with self.subTest(method=method):
        # Expected results are computed using TFP Beta-to-Beta KL divergence.
        expected_result_1 = getattr(tfd.Beta(alpha1, beta1), method)(tfp_dist2)
        expected_result_2 = getattr(tfp_dist2, method)(tfd.Beta(alpha1, beta1))
        if mode_string == 'distrax_to_distrax':
          result1 = self.variant(getattr(distrax_dist1, method))(distrax_dist2)
          result2 = self.variant(getattr(distrax_dist2, method))(distrax_dist1)
        elif mode_string == 'distrax_to_tfp':
          result1 = self.variant(getattr(distrax_dist1, method))(tfp_dist2)
          result2 = self.variant(getattr(distrax_dist2, method))(tfp_dist1)
        elif mode_string == 'tfp_to_distrax':
          result1 = self.variant(getattr(tfp_dist1, method))(distrax_dist2)
          result2 = self.variant(getattr(tfp_dist2, method))(distrax_dist1)
        else:
          raise ValueError(f'Unsupported mode: {mode_string}')
        self.assertion_fn(rtol=3e-2)(result1, expected_result_1)
        self.assertion_fn(rtol=3e-2)(result2, expected_result_2)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('dirichlet_dirichlet', dirichlet.Dirichlet),
      ('beta_dirichlet', beta.Beta),
  )
  def test_kl_raises_on_wrong_dims(self, dist2_type):
    rng = np.random.default_rng(42)
    # Distribution 1 is Dirichlet.
    concentration1 = np.abs(rng.normal(size=(5, 3)))
    dist1 = self.distrax_cls(concentration1)
    # Distribution 2 is either Dirichlet or Beta.
    if dist2_type is dirichlet.Dirichlet:
      dist2_kwargs = {'concentration': rng.uniform(size=(5, 4))}
    elif dist2_type is beta.Beta:
      dist2_kwargs = {'alpha': rng.uniform(size=(5,)),
                      'beta': rng.uniform(size=(5,))}
    else:
      raise ValueError(f'Unsupported distribution: {dist2_type}')
    dist2 = dist2_type(**dist2_kwargs)
    with self.assertRaises(ValueError):
      self.variant(dist1.kl_divergence)(dist2)
    with self.assertRaises(ValueError):
      self.variant(dist2.kl_divergence)(dist1)

  def test_jitable(self):
    rng = np.random.default_rng(42)
    concentration = np.abs(rng.normal(size=(4,)))
    super()._test_jittable(
        (concentration,), assertion_fn=self.assertion_fn(rtol=1e-3))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    rng = np.random.default_rng(42)
    concentration = np.abs(rng.normal(size=(6, 5, 4)))
    dist = self.distrax_cls(concentration)
    self.assertIsInstance(dist, self.distrax_cls)
    self.assertion_fn(rtol=1e-3)(
        dist[slice_].concentration, concentration[slice_])

  def test_slice_ellipsis(self):
    rng = np.random.default_rng(42)
    concentration = np.abs(rng.normal(size=(6, 5, 4)))
    dist = self.distrax_cls(concentration)
    self.assertIsInstance(dist, self.distrax_cls)
    self.assertion_fn(rtol=1e-3)(
        dist[..., -1].concentration, concentration[:, -1, :])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `mvn_full_covariance.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions.mvn_full_covariance import MultivariateNormalFullCovariance
from distrax._src.utils import equivalence
import jax.experimental
import jax.numpy as jnp
import numpy as np


def _sample_covariance_matrix(rng, shape):
  matrix = rng.normal(size=shape)
  matrix_t = np.vectorize(np.transpose, signature='(k,k)->(k,k)')(matrix)
  return np.matmul(matrix, matrix_t)


class MultivariateNormalFullCovarianceTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(MultivariateNormalFullCovariance)

  @parameterized.named_parameters(
      ('all inputs are None', {}),
      ('wrong dimension of loc', {
          'loc': np.array(0.),
      }),
      ('covariance_matrix is 0d', {
          'covariance_matrix': np.array(1.),
      }),
      ('covariance_matrix is 1d', {
          'covariance_matrix': np.ones((4,)),
      }),
      ('covariance_matrix is not square', {
          'covariance_matrix': np.ones((4, 3)),
      }),
      ('inconsistent loc and covariance_matrix', {
          'loc': np.zeros((4,)),
          'covariance_matrix': np.eye(5),
      }),
  )
  def test_raises_on_wrong_inputs(self, dist_kwargs):
    with self.assertRaises(ValueError):
      self.distrax_cls(**dist_kwargs)

  @parameterized.named_parameters(
      ('loc provided', {'loc': np.zeros((4,))}),
      ('covariance_matrix provided', {'covariance_matrix': np.eye(4)}),
  )
  def test_default_properties(self, dist_kwargs):
    dist = self.distrax_cls(**dist_kwargs)
    self.assertion_fn(rtol=1e-3)(dist.loc, jnp.zeros((4,)))
    self.assertion_fn(rtol=1e-3)(dist.covariance_matrix, jnp.eye(4))

  @parameterized.named_parameters(
      ('unbatched', (), (4,), (4, 4)),
      ('batched loc', (7,), (7, 4), (4, 4)),
      ('batched covariance_matrix', (7,), (4,), (7, 4, 4)),
  )
  def test_properties(self, batch_shape, loc_shape, covariance_matrix_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    covariance_matrix = _sample_covariance_matrix(rng, covariance_matrix_shape)
    dist = self.distrax_cls(loc=loc, covariance_matrix=covariance_matrix)
    self.assertEqual(dist.batch_shape, batch_shape)
    self.assertion_fn(rtol=1e-3)(
        dist.loc, jnp.broadcast_to(loc, batch_shape + (4,)))
    self.assertion_fn(rtol=1e-3)(dist.covariance_matrix, jnp.broadcast_to(
        covariance_matrix, batch_shape + (4, 4)))

  @chex.all_variants
  @parameterized.named_parameters(
      ('unbatched, no shape', (), (4,), (4, 4)),
      ('batched loc, no shape', (), (7, 4), (4, 4)),
      ('batched covariance_matrix, no shape', (), (4,), (7, 4, 4)),
      ('unbatched, with shape', (3,), (4,), (4, 4)),
      ('batched loc, with shape', (3,), (7, 4), (4, 4)),
      ('batched covariance_matrix, with shape', (3,), (4,), (7, 4, 4)),
  )
  def test_sample_shape(self, sample_shape, loc_shape, covariance_matrix_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    covariance_matrix = _sample_covariance_matrix(rng, covariance_matrix_shape)
    dist_kwargs = {'loc': loc, 'covariance_matrix': covariance_matrix}
    super()._test_sample_shape(
        dist_args=(), dist_kwargs=dist_kwargs, sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist_params = {
          'loc': np.array([0., 0.], dtype),
          'covariance_matrix': np.array([[1., 0.], [0., 1.]], dtype)}
      dist = self.distrax_cls(**dist_params)
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('unbatched, unbatched value', (4,), (4,), (4, 4)),
      ('batched loc, unbatched value', (4,), (7, 4), (4, 4)),
      ('batched covariance_matrix, unbatched value', (4,), (4,), (7, 4, 4)),
      ('unbatched, batched value', (3, 7, 4), (4,), (4, 4)),
      ('batched loc, batched value', (3, 7, 4), (7, 4), (4, 4)),
      ('batched covariance_matrix, batched value', (3, 7, 4), (4,), (7, 4, 4)),
  )
  def test_log_prob(self, value_shape, loc_shape, covariance_matrix_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    covariance_matrix = _sample_covariance_matrix(rng, covariance_matrix_shape)
    dist_kwargs = {'loc': loc, 'covariance_matrix': covariance_matrix}
    value = rng.normal(size=value_shape)
    super()._test_attribute(
        attribute_string='log_prob',
        dist_kwargs=dist_kwargs,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('unbatched', (4,), (4, 4)),
      ('batched loc', (7, 4), (4, 4)),
      ('batched covariance_matrix', (4,), (7, 4, 4)),
  )
  def test_method(self, loc_shape, covariance_matrix_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    covariance_matrix = _sample_covariance_matrix(rng, covariance_matrix_shape)
    dist_kwargs = {'loc': loc, 'covariance_matrix': covariance_matrix}
    for method in ['entropy', 'mean', 'stddev', 'variance',
                   'covariance', 'mode']:
      if method == 'covariance':
        rtol = 2e-2
      elif method in ['stddev', 'variance']:
        rtol = 6e-3
      else:
        rtol = 1e-3
      with self.subTest(method=method):
        super()._test_attribute(
            method,
            dist_kwargs=dist_kwargs,
            assertion_fn=self.assertion_fn(rtol=rtol))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    rng = np.random.default_rng(2022)
    dist1_kwargs = {
        'loc': rng.normal(size=(5, 1, 4)),
        'covariance_matrix': _sample_covariance_matrix(rng, (3, 4, 4)),
    }
    dist2_kwargs = {
        'loc': rng.normal(size=(3, 4)),
        'covariance_matrix': _sample_covariance_matrix(rng, (4, 4)),
    }
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs=dist1_kwargs,
        dist2_kwargs=dist2_kwargs,
        assertion_fn=self.assertion_fn(rtol=1e-3))

  def test_jittable(self):
    super()._test_jittable(
        dist_kwargs={'loc': np.zeros((4,))},
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=(6, 5, 4))
    covariance_matrix = _sample_covariance_matrix(rng, (4, 4))
    dist_kwargs = {'loc': loc, 'covariance_matrix': covariance_matrix}
    dist = self.distrax_cls(**dist_kwargs)
    self.assertEqual(dist[slice_].batch_shape, loc[slice_].shape[:-1])
    self.assertEqual(dist[slice_].event_shape, dist.event_shape)
    self.assertion_fn(rtol=1e-3)(dist[slice_].mean(), loc[slice_])

  def test_slice_ellipsis(self):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=(6, 5, 4))
    covariance_matrix = _sample_covariance_matrix(rng, (4, 4))
    dist_kwargs = {'loc': loc, 'covariance_matrix': covariance_matrix}
    dist = self.distrax_cls(**dist_kwargs)
    self.assertEqual(dist[..., -1].batch_shape, (6,))
    self.assertEqual(dist[..., -1].event_shape, dist.event_shape)
    self.assertion_fn(rtol=1e-3)(dist[..., -1].mean(), loc[:, -1, :])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `tfp_compatible_bijector.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors.block import Block
from distrax._src.bijectors.chain import Chain
from distrax._src.bijectors.lambda_bijector import Lambda
from distrax._src.bijectors.scalar_affine import ScalarAffine
from distrax._src.bijectors.tanh import Tanh
from distrax._src.bijectors.tfp_compatible_bijector import tfp_compatible_bijector
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfb = tfp.bijectors
tfd = tfp.distributions

RTOL = 3e-3


class TFPCompatibleBijectorTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('lambda identity, 0d sample', lambda: Block(lambda x: x, 1),
       tfb.Identity, ()),
      ('lambda identity, 1d sample', lambda: Block(lambda x: x, 1),
       tfb.Identity, (5,)),
      ('lambda identity, 2d sample', lambda: Block(lambda x: x, 1),
       tfb.Identity, (7, 5)),
      ('ScalarAffine, 0d sample',
       lambda: Block(ScalarAffine(3, 2), 1),
       lambda: tfb.Chain([tfb.Shift(3), tfb.Scale(2)]), ()),
      ('ScalarAffine, 1d sample',
       lambda: Block(ScalarAffine(3, 2), 1),
       lambda: tfb.Chain([tfb.Shift(3), tfb.Scale(2)]), (5,)),
      ('ScalarAffine, 2d sample',
       lambda: Block(ScalarAffine(3, 2), 1),
       lambda: tfb.Chain([tfb.Shift(3), tfb.Scale(2)]), (7, 5)),
      ('Tanh, 0d sample', lambda: Block(Tanh(), 1), tfb.Tanh, ()),
      ('Tanh, 1d sample', lambda: Block(Tanh(), 1), tfb.Tanh, (5,)),
      ('Tanh, 2d sample',
       lambda: Block(Tanh(), 1), tfb.Tanh, (7, 5)),
      ('Chain(Tanh, ScalarAffine), 0d sample',
       lambda: Block(Chain([Tanh(), ScalarAffine(0.2, 1.2)]), 1),
       lambda: tfb.Chain([tfb.Tanh(), tfb.Shift(0.2), tfb.Scale(1.2)]), ()),
      ('Chain(Tanh, ScalarAffine), 1d sample',
       lambda: Block(Chain([Tanh(), ScalarAffine(0.2, 1.2)]), 1),
       lambda: tfb.Chain([tfb.Tanh(), tfb.Shift(0.2), tfb.Scale(1.2)]), (5,)),
      ('Chain(Tanh, ScalarAffine), 2d sample',
       lambda: Block(Chain([Tanh(), ScalarAffine(0.2, 1.2)]), 1),
       lambda: tfb.Chain([tfb.Tanh(), tfb.Shift(0.2), tfb.Scale(1.2)]), (3, 5)),
  )
  def test_transformed_distribution(
      self, dx_bijector_fn, tfp_bijector_fn, sample_shape):
    base_dist = tfd.MultivariateNormalDiag(np.zeros((3, 2)), np.ones((3, 2)))
    dx_bijector = dx_bijector_fn()
    wrapped_bijector = tfp_compatible_bijector(dx_bijector)
    tfp_bijector = tfp_bijector_fn()
    dist_with_wrapped = tfd.TransformedDistribution(base_dist, wrapped_bijector)
    dist_tfp_only = tfd.TransformedDistribution(base_dist, tfp_bijector)

    with self.subTest('sample'):
      dist_with_wrapped.sample(
          seed=jax.random.PRNGKey(0), sample_shape=sample_shape)

    with self.subTest('log_prob'):
      y = dist_tfp_only.sample(
          seed=jax.random.PRNGKey(0), sample_shape=sample_shape)
      log_prob_wrapped = dist_with_wrapped.log_prob(y)
      log_prob_tfp_only = dist_tfp_only.log_prob(y)
      np.testing.assert_allclose(log_prob_wrapped, log_prob_tfp_only, rtol=RTOL)

  @parameterized.named_parameters(
      ('identity', lambda: Lambda(lambda x: x, is_constant_jacobian=True),
       tfb.Identity, np.array([1, 1.5, 2], dtype=np.float32)),
      ('affine', lambda: ScalarAffine(np.ones(3, dtype=np.float32),  # pylint: disable=g-long-lambda
                                      np.full(3, 5.5, dtype=np.float32)),
       lambda: tfb.Chain([tfb.Shift(np.ones(3, dtype=np.float32)),  # pylint: disable=g-long-lambda
                          tfb.Scale(np.full(3, 5.5, dtype=np.float32))]),
       np.array([1, 1.5, 2], dtype=np.float32)),
      ('tanh', Tanh, tfb.Tanh, np.array([-0.1, 0.01, 0.1], dtype=np.float32)),
      ('chain(tanh, affine)', lambda: Chain([Tanh(), ScalarAffine(0.2, 1.2)]),
       lambda: tfb.Chain([tfb.Tanh(), tfb.Shift(0.2), tfb.Scale(1.2)]),
       np.array([-0.1, 0.01, 0.1], dtype=np.float32))
  )
  def test_chain(self, dx_bijector_fn, tfb_bijector_fn, event):
    dx_bij = tfp_compatible_bijector(dx_bijector_fn())
    tfp_bij = tfb_bijector_fn()

    chain_with_dx = tfb.Chain([tfb.Shift(1.0), tfb.Scale(3.0), dx_bij])
    chain_with_tfp = tfb.Chain([tfb.Shift(1.0), tfb.Scale(3.0), tfp_bij])

    with self.subTest('forward'):
      y_dx = chain_with_dx.forward(event)
      y_tfp = chain_with_tfp.forward(event)
      np.testing.assert_allclose(y_dx, y_tfp, rtol=RTOL)

    with self.subTest('inverse'):
      y = chain_with_tfp.forward(event)
      x_dx = chain_with_dx.inverse(y)
      np.testing.assert_allclose(x_dx, event, rtol=RTOL)

  @parameterized.named_parameters(
      ('identity', lambda: Lambda(lambda x: x, is_constant_jacobian=True),
       tfb.Identity, np.array([1, 1.5, 2], dtype=np.float32)),
      ('affine', lambda: ScalarAffine(np.ones(3, dtype=np.float32),  # pylint: disable=g-long-lambda
                                      np.full(3, 5.5, dtype=np.float32)),
       lambda: tfb.Chain([tfb.Shift(np.ones(3, dtype=np.float32)),  # pylint: disable=g-long-lambda
                          tfb.Scale(np.full(3, 5.5, dtype=np.float32))]),
       np.array([1, 1.5, 2], dtype=np.float32)),
      ('tanh', Tanh, tfb.Tanh, np.array([-0.1, 0.01, 0.1], dtype=np.float32)),
      ('chain(tanh, affine)', lambda: Chain([Tanh(), ScalarAffine(0.2, 1.2)]),
       lambda: tfb.Chain([tfb.Tanh(), tfb.Shift(0.2), tfb.Scale(1.2)]),
       np.array([-0.1, 0.01, 0.1], dtype=np.float32)),
  )
  def test_invert(self, dx_bijector_fn, tfb_bijector_fn, event):
    dx_bij = tfp_compatible_bijector(dx_bijector_fn())
    tfp_bij = tfb_bijector_fn()

    invert_with_dx = tfb.Invert(dx_bij)
    invert_with_tfp = tfb.Invert(tfp_bij)

    with self.subTest('forward'):
      y_dx = invert_with_dx.forward(event)
      y_tfp = invert_with_tfp.forward(event)
      np.testing.assert_allclose(y_dx, y_tfp, rtol=RTOL)

    with self.subTest('inverse'):
      y = invert_with_tfp.forward(event)
      x_dx = invert_with_dx.inverse(y)
      np.testing.assert_allclose(x_dx, event, rtol=RTOL)

  @parameterized.named_parameters(
      ('identity', lambda: Lambda(lambda x: x, is_constant_jacobian=True),
       tfb.Identity, np.array([1, 1.5, 2], dtype=np.float32)),
      ('affine', lambda: ScalarAffine(np.ones(3, dtype=np.float32),  # pylint: disable=g-long-lambda
                                      np.full(3, 5.5, dtype=np.float32)),
       lambda: tfb.Chain([tfb.Shift(np.ones(3, dtype=np.float32)),  # pylint: disable=g-long-lambda
                          tfb.Scale(np.full(3, 5.5, dtype=np.float32))]),
       np.array([1, 1.5, 2], dtype=np.float32)),
      ('tanh', Tanh, tfb.Tanh, np.array([-0.1, 0.01, 0.1], dtype=np.float32)),
      ('chain(tanh, affine)', lambda: Chain([Tanh(), ScalarAffine(0.2, 1.2)]),
       lambda: tfb.Chain([tfb.Tanh(), tfb.Shift(0.2), tfb.Scale(1.2)]),
       np.array([-0.1, 0.01, 0.1], dtype=np.float32))
  )
  def test_forward_and_inverse(
      self, dx_bijector_fn, tfp_bijector_fn, event):
    dx_bij = tfp_compatible_bijector(dx_bijector_fn())
    tfp_bij = tfp_bijector_fn()

    with self.subTest('forward'):
      dx_out = dx_bij.forward(event)
      tfp_out = tfp_bij.forward(event)
      np.testing.assert_allclose(dx_out, tfp_out, rtol=RTOL)

    with self.subTest('inverse'):
      y = tfp_bij.forward(event)
      dx_out = dx_bij.inverse(y)
      tfp_out = tfp_bij.inverse(y)
      np.testing.assert_allclose(dx_out, tfp_out, rtol=RTOL)

  @parameterized.named_parameters(
      ('identity', lambda: Lambda(lambda x: x, is_constant_jacobian=True),
       tfb.Identity, np.array([1, 1.5, 2], dtype=np.float32)),
      ('affine', lambda: ScalarAffine(np.ones(3, dtype=np.float32),  # pylint: disable=g-long-lambda
                                      np.full(3, 5.5, dtype=np.float32)),
       lambda: tfb.Chain([tfb.Shift(np.ones(3, dtype=np.float32)),  # pylint: disable=g-long-lambda
                          tfb.Scale(np.full(3, 5.5, dtype=np.float32))]),
       np.array([1, 1.5, 2], dtype=np.float32)),
      ('tanh', Tanh, tfb.Tanh, np.array([-0.1, 0.01, 0.1], dtype=np.float32)),
      ('chain(tanh, affine)', lambda: Chain([Tanh(), ScalarAffine(0.2, 1.2)]),
       lambda: tfb.Chain([tfb.Tanh(), tfb.Shift(0.2), tfb.Scale(1.2)]),
       np.array([-0.1, 0.01, 0.1], dtype=np.float32))
  )
  def test_log_det_jacobian(self, dx_bijector_fn, tfp_bijector_fn, event):
    base_bij = dx_bijector_fn()
    dx_bij = tfp_compatible_bijector(base_bij)
    tfp_bij = tfp_bijector_fn()

    with self.subTest('forward'):
      dx_out = dx_bij.forward_log_det_jacobian(
          event, event_ndims=base_bij.event_ndims_in)
      tfp_out = tfp_bij.forward_log_det_jacobian(
          event, event_ndims=base_bij.event_ndims_in)
      np.testing.assert_allclose(dx_out, tfp_out, rtol=RTOL)

    with self.subTest('inverse'):
      y = tfp_bij.forward(event)
      dx_out = dx_bij.inverse_log_det_jacobian(
          y, event_ndims=base_bij.event_ndims_out)
      tfp_out = tfp_bij.inverse_log_det_jacobian(
          y, event_ndims=base_bij.event_ndims_out)
      np.testing.assert_allclose(dx_out, tfp_out, rtol=RTOL)

    with self.subTest('experimental_compute_density_correction'):
      dx_out = dx_bij.forward_log_det_jacobian(
          event, event_ndims=base_bij.event_ndims_in)
      dx_dcorr_out, space = dx_bij.experimental_compute_density_correction(
          event, tangent_space=tfp.experimental.tangent_spaces.FullSpace(),
          event_ndims=base_bij.event_ndims_in)
      np.testing.assert_allclose(dx_out, dx_dcorr_out, rtol=RTOL)
      self.assertIsInstance(space, tfp.experimental.tangent_spaces.FullSpace)

  @parameterized.named_parameters(
      ('identity unbatched',
       lambda: Lambda(lambda x: x, is_constant_jacobian=True), ()),
      ('identity 1d-batch',
       lambda: Lambda(lambda x: x, is_constant_jacobian=True), (3,)),
      ('identity 2d-batch',
       lambda: Lambda(lambda x: x, is_constant_jacobian=True), (5, 3)),
      ('affine unbatched', lambda: ScalarAffine(1.0, 5.5), ()),
      ('affine 1d-batch', lambda: ScalarAffine(1.0, 5.5), (3,)),
      ('affine 2d-batch', lambda: ScalarAffine(1.0, 5.5), (5, 3)),
      ('tanh unbatched', Tanh, ()),
      ('tanh 1d-batch', Tanh, (3,)),
      ('tanh 2d-batch', Tanh, (5, 3)),
      ('chain(tanh, affine) unbatched',
       lambda: Chain([Tanh(), ScalarAffine(0.2, 1.2)]), ()),
      ('chain(tanh, affine) 1d-batch',
       lambda: Chain([Tanh(), ScalarAffine(0.2, 1.2)]), (3,)),
      ('chain(tanh, affine) 2d-batch',
       lambda: Chain([Tanh(), ScalarAffine(0.2, 1.2)]), (5, 3)),
  )
  def test_batched_events(self, bij_fn, batch_shape):
    base = tfd.MultivariateNormalDiag(np.zeros(batch_shape + (3,)),
                                      np.ones(batch_shape + (3,)))
    bij = tfp_compatible_bijector(bij_fn())
    dist = tfd.TransformedDistribution(base, bij)

    with self.subTest('sample'):
      sample = dist.sample(seed=jax.random.PRNGKey(0))
      chex.assert_shape(sample, batch_shape + (3,))

    with self.subTest('log_prob'):
      sample = dist.sample(seed=jax.random.PRNGKey(0))
      log_prob = dist.log_prob(sample)
      chex.assert_shape(log_prob, batch_shape)

  def test_with_different_event_ndims(self):
    dx_bij = Lambda(forward=lambda x: x.reshape(x.shape[:-1] + (2, 3)),
                    inverse=lambda y: y.reshape(y.shape[:-2] + (6,)),
                    forward_log_det_jacobian=lambda _: 0,
                    inverse_log_det_jacobian=lambda _: 0,
                    is_constant_jacobian=True,
                    event_ndims_in=1, event_ndims_out=2)
    tfp_bij = tfp_compatible_bijector(dx_bij)

    with self.subTest('forward_event_ndims'):
      assert tfp_bij.forward_event_ndims(1) == 2
      assert tfp_bij.forward_event_ndims(2) == 3

    with self.subTest('inverse_event_ndims'):
      assert tfp_bij.inverse_event_ndims(2) == 1
      assert tfp_bij.inverse_event_ndims(3) == 2

    with self.subTest('forward_event_ndims with incorrect input'):
      with self.assertRaises(ValueError):
        tfp_bij.forward_event_ndims(0)

    with self.subTest('inverse_event_ndims with incorrect input'):
      with self.assertRaises(ValueError):
        tfp_bij.inverse_event_ndims(0)

      with self.assertRaises(ValueError):
        tfp_bij.inverse_event_ndims(1)

    with self.subTest('forward_event_shape'):
      y_shape = tfp_bij.forward_event_shape((6,))
      y_shape_tensor = tfp_bij.forward_event_shape_tensor((6,))
      self.assertEqual(y_shape, (2, 3))
      np.testing.assert_array_equal(y_shape_tensor, jnp.array((2, 3)))

    with self.subTest('inverse_event_shape'):
      x_shape = tfp_bij.inverse_event_shape((2, 3))
      x_shape_tensor = tfp_bij.inverse_event_shape_tensor((2, 3))
      self.assertEqual(x_shape, (6,))
      np.testing.assert_array_equal(x_shape_tensor, jnp.array((6,)))

    with self.subTest('TransformedDistribution with correct event_ndims'):
      base = tfd.MultivariateNormalDiag(np.zeros(6), np.ones(6))
      dist = tfd.TransformedDistribution(base, tfp_bij)
      chex.assert_equal(dist.event_shape, (2, 3))

      sample = dist.sample(seed=jax.random.PRNGKey(0))
      chex.assert_shape(sample, (2, 3))

      log_prob = dist.log_prob(sample)
      chex.assert_shape(log_prob, ())

    with self.subTest('TransformedDistribution with incorrect event_ndims'):
      base = tfd.Normal(np.zeros(6), np.ones(6))
      dist = tfd.TransformedDistribution(base, tfp_bij)
      with self.assertRaises(ValueError):
        _ = dist.event_shape

if __name__ == '__main__':
  absltest.main()




# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Chain Bijector for composing a sequence of Bijector transformations."""

from typing import List, Sequence, Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.utils import conversion

Array = base.Array
BijectorLike = base.BijectorLike
BijectorT = base.BijectorT


class Chain(base.Bijector):
  """Composition of a sequence of bijectors into a single bijector.

  Bijectors are composable: if `f` and `g` are bijectors, then `g o f` is also
  a bijector. Given a sequence of bijectors `[f1, ..., fN]`, this class
  implements the bijector defined by `fN o ... o f1`.

  NOTE: the bijectors are applied in reverse order from the order they appear in
  the sequence. For example, consider the following code where `f` and `g` are
  two bijectors:
  ```
  layers = []
  layers.append(f)
  layers.append(g)
  bijector = distrax.Chain(layers)
  y = bijector.forward(x)
  ```
  The above code will transform `x` by first applying `g`, then `f`, so that
  `y = f(g(x))`.
  """

  def __init__(self, bijectors: Sequence[BijectorLike]):
    """Initializes a Chain bijector.

    Args:
      bijectors: a sequence of bijectors to be composed into one. Each bijector
        can be a distrax bijector, a TFP bijector, or a callable to be wrapped
        by `Lambda`. The sequence must contain at least one bijector.
    """
    if not bijectors:
      raise ValueError("The sequence of bijectors cannot be empty.")
    self._bijectors = [conversion.as_bijector(b) for b in bijectors]

    # Check that neighboring bijectors in the chain have compatible dimensions
    for i, (outer, inner) in enumerate(zip(self._bijectors[:-1],
                                           self._bijectors[1:])):
      if outer.event_ndims_in != inner.event_ndims_out:
        raise ValueError(
            f"The chain of bijector event shapes are incompatible. Bijector "
            f"{i} ({outer.name}) expects events with {outer.event_ndims_in} "
            f"dimensions, while Bijector {i+1} ({inner.name}) produces events "
            f"with {inner.event_ndims_out} dimensions.")

    is_constant_jacobian = all(b.is_constant_jacobian for b in self._bijectors)
    is_constant_log_det = all(b.is_constant_log_det for b in self._bijectors)
    super().__init__(
        event_ndims_in=self._bijectors[-1].event_ndims_in,
        event_ndims_out=self._bijectors[0].event_ndims_out,
        is_constant_jacobian=is_constant_jacobian,
        is_constant_log_det=is_constant_log_det)

  @property
  def bijectors(self) -> List[BijectorT]:
    """The list of bijectors in the chain."""
    return self._bijectors

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    for bijector in reversed(self._bijectors):
      x = bijector.forward(x)
    return x

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    for bijector in self._bijectors:
      y = bijector.inverse(y)
    return y

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    x, log_det = self._bijectors[-1].forward_and_log_det(x)
    for bijector in reversed(self._bijectors[:-1]):
      x, ld = bijector.forward_and_log_det(x)
      log_det += ld
    return x, log_det

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    y, log_det = self._bijectors[0].inverse_and_log_det(y)
    for bijector in self._bijectors[1:]:
      y, ld = bijector.inverse_and_log_det(y)
      log_det += ld
    return y, log_det

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is Chain:  # pylint: disable=unidiomatic-typecheck
      if len(self.bijectors) != len(other.bijectors):
        return False
      for bij1, bij2 in zip(self.bijectors, other.bijectors):
        if not bij1.same_as(bij2):
          return False
      return True
    elif len(self.bijectors) == 1:
      return self.bijectors[0].same_as(other)

    return False

